# coding=utf-8
from .kernel import Kernel
from ..utils import DEFAULT_KERNEL_TYPE


def factory(kernel_type=DEFAULT_KERNEL_TYPE, **kwargs) -> Kernel:
    r"""
    Creates a kernel based on the specified name with the specified arguments. This is the same as
    calling `kerch.kernel.name(*args, **kwargs)` (if `name` is not a string here). This allows for the creation of kernel where
    the name of kernel is passed as a string.

    :param kernel_type: Type of kernel chosen. For the possible choices, please refer to the `Factory Type` column of the
        :doc:`../kernel/index` documentation. Defaults to :py:data:`kerch.DEFAULT_KERNEL_TYPE`.
    :param \**kwargs: Arguments to be passed to the kernel constructor, such as `sample` or `sigma`. If an argument is
        passed that does not exist (e.g. `sigma` to a `linear` kernel), it will just be neglected. For the default
        values, please refer to the default values of the requested kernel.
    :type kernel_type: str, optional
    :type \**kwargs: dict, optional
    :return: An instantiation of the specified kernel.
    """

    kernel = class_factory(kernel_type)
    return kernel(**kwargs)


def class_factory(kernel_type=DEFAULT_KERNEL_TYPE):
    def case_insensitive_getattr(obj, attr):
        for a in dir(obj):
            if a.lower() == attr.lower():
                return getattr(obj, a)
        return None

    import kerch.kernel
    kernel = case_insensitive_getattr(kerch.kernel, kernel_type.replace("_", ""))
    if kernel is None:
        raise NameError("Invalid kernel type.")
    return kernel




# coding=utf-8
import torch
from ..utils import castf, DEFAULT_KERNEL_TYPE
from ..feature.logger import _GLOBAL_LOGGER


@torch.no_grad()
def knn(dists: torch.Tensor, observations: torch.Tensor, num: int = 1) -> torch.Tensor:
    r"""
    For each distance ``dists``, returns the average of the ``num`` smallest corresponding observations.

    :param dists: coefficients used in the knn.
    :type dists: torch.Tensor [num_points, num_observations]
    :param observations: observation corresponding to each weight dimension.
    :type observations: torch.Tensor [num_observations, dim_observations]
    :param num: number of nearest neighbors. Defaults to 1.
    :type num: int, optional
    :return: KNN
    :rtype: torch.Tensor [num_points, dim_observations]
    """

    # PRELIMINARIES
    dists = castf(dists)
    observations = castf(observations)

    num_points, num_coefficients = dists.shape
    num_observations = observations.shape[0]

    # DEFENSIVE
    try:
        num = int(num)
    except ValueError:
        raise ValueError('The argument num is not an integer.')

    assert num_coefficients == num_observations, \
        f'KNN: Incorrect number of coefficients ({num_coefficients}), ' \
        f'compared to the number of points ({num_observations}).'

    assert num <= num_coefficients, \
        (f"Too much required neighbors ({num}) compared to the number of observations points "
         f"({num_observations}). Please insure that the argument num is not greater than the number of observations "
         f"points.")

    assert num > 0, \
        f"The number of required neighbors num must be strictly positive ({num})."

    # PRE-IMAGE
    if dists.min() >= 0:
        _GLOBAL_LOGGER._logger.warning('There are negative distances for kNN. The coefficients are changed.')
        dists = dists - dists.min()
    _, indices = torch.topk(-dists, k=num, dim=1)
    kept_sample = observations[indices]
    return torch.mean(kept_sample, dim=1)


@torch.no_grad()
def kernel_knn(domain: torch.Tensor, observations: torch.Tensor, num: int = 1, kernel_type: str = DEFAULT_KERNEL_TYPE,
               **kwargs) -> torch.Tensor:
    r"""
    For each coefficient, returns the average of the ``num`` greatest corresponding kernel values on the domain.
    The kernel is defined as in :py:func:`kerch.kernel.factory`.

    :param domain: domain corresponding to each observation.
    :type domain: torch.Tensor [num_observations, dim_domain]
    :param observations: observation corresponding to each domain entry.
    :type observations: torch.Tensor [num_observations, dim_observations]
    :param num: number of nearest neighbors. Defaults to 1.
    :type num: int, optional
    :param kernel_type: Type of kernel chosen. For the possible choices, please refer to the `Factory Type` column of the
        :doc:`../kernel/index` documentation. Defaults to :py:data:`kerch.DEFAULT_KERNEL_TYPE`.
    :param \**kwargs: Arguments to be passed to the kernel constructor, such as `sample` or `sigma`. If an argument is
        passed that does not exist (e.g. `sigma` to a `linear` kernel), it will just be neglected. For the default
        values, please refer to the default values of the requested kernel.
    :type kernel_type: str, optional
    :type \**kwargs: dict, optional
    :return: KNN
    :rtype: torch.Tensor [num_points, dim_observations]
    """
    domain = castf(domain)
    observations = castf(observations)

    assert domain.shape[0] == observations.shape[
        0], f"Not the same number of domain {domain.shape[0]} and coefficients points {domain.shape[0]}."

    from ..kernel import factory

    k = factory(kernel_type=kernel_type, sample=domain, **kwargs)
    return knn(dists=-k.K, observations=observations, num=num)



"""
Abstract RKM View class.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021, rewritten in June 2022
"""

import torch
from torch import Tensor
from abc import ABCMeta, abstractmethod
from typing import Union

from kerch import utils
from ..feature.stochastic import Stochastic


@utils.extend_docstring(Stochastic)
class _View(Stochastic, metaclass=ABCMeta):
    r"""
    :param dim_output: Output dimension. If None, it will later be assigned by the target is relevant. Defaults to None.
    :param representation: Specifies if the level works in primal or dual representation. The dual representation
        ('dual') is guaranteed to always have a finite dimensional formulation (the RKHS is finite dimensional as the
        sample has a finite number of datapoints). If the primal formulation ('primal') is available, the number of
        parameters (proportional to the property `dim_feature`) will potentially be smaller than for the dual
        (proportional to the number of sample datapoints, hence the formulation will be lighter hence faster.
        Defaults to 'dual'.
    :param weight: Weight values to start with. This is most of the cases not necessary if the level is meant to be
        trained based on a gradient or fitted. Defaults to None.
    :param hidden: Hidden values to start with. This is most of the cases not necessary if the level is meant to be
        trained based on a gradient or fitted. Defaults to None.
    :param_trainable: Boolean specifying whether the model parameters (weight, hidden and bias if applicable) are meant
        to have a gradient available. This is relevant is the level is to be trained based on a gradient. This is
        irrelevant if the model is meant to be fitted by a linear system. Defaults to False.

    :type dim_output: int, optional
    :type representation: str, optional
    :type weight: torch.Tensor [dim_feature, dim_output], optional
    :type hidden: torch.Tensor [num_sample, dim_output], optional
    :param_trainable: bool, optional
    """

    def __init__(self, *args, **kwargs):
        """
        A View is made of a kernel and primal or dual variables. This second part is handled by the daughter classes.
        """
        super(_View, self).__init__(*args, **kwargs)
        self._dim_output = kwargs.pop('dim_output', None)
        self._representation = utils.check_representation(kwargs.pop('representation', 'dual'), cls=self)

        weight = kwargs.pop('weight', None)
        hidden = kwargs.pop('hidden', None)

        # INITIATE
        self._level_trainable = kwargs.pop('level_trainable', False)
        self._dual_param = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE), self._level_trainable)
        self._primal_param = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE), self._level_trainable)
        if weight is not None and hidden is not None:
            self._logger.info("Both the hidden and the weight are set. Priority is given to the weight values.")
            self.primal_param = weight
        elif weight is None:
            self.dual_param = hidden
        elif hidden is None:
            self.primal_param = weight

        self._attached_primal_param = None

    @property
    def hparams_fixed(self) -> dict:
        constraint = 'soft' if self._level_trainable else 'hard'
        return {'Output dimension': self.dim_output,
                'Representation': self.representation,
                'Parameters trainable': self._level_trainable,
                'Constraint': constraint,
                **super(_View, self).hparams_fixed}

    def _reset_dual(self) -> None:
        self._dual_param = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE,
                                                          device=self._param_device),
                                              requires_grad=self._dual_param.requires_grad)

    @abstractmethod
    def _reset_primal(self) -> None:
        pass

    def _init_dual(self) -> None:
        assert self._num_total is not None, "No data has been initialized yet."
        assert self._dim_output is not None, "No output dimension has been provided."
        self.dual_param = torch.nn.init.orthogonal_(torch.empty((self._num_total, self.dim_output),
                                                                dtype=utils.FTYPE, device=self._param_device))
        self._logger.info('The hidden variables are initialized.')

    def _init_primal(self) -> None:
        assert self._dim_output is not None, "No output dimension has been provided."
        self.primal_param = torch.nn.init.orthogonal_(torch.empty((self.dim_feature, self.dim_output),
                                                                  dtype=utils.FTYPE, device=self._param_device))
        self._logger.info('The weights are initialized.')

    def init_parameters(self, representation=None, overwrite=True) -> None:
        """
        Initializes the model parameters: the weight in primal and the hidden values in dual.
        This is suitable for gradient-based training.

        :param representation: 'primal' or 'dual'
        :param overwrite: Does not initialize already initialized parameters if False., defaults to True
        :type representation: str, optional
        :type overwrite: bool, optional
        """
        representation = utils.check_representation(representation, self._representation, cls=self)

        def init_weight():
            if overwrite or not self._primal_param_exists:
                self._init_primal()

        def init_hidden():
            if overwrite or not self._dual_param_exists:
                self._init_dual()

        switcher = {"primal": init_weight,
                    "dual": init_hidden}
        switcher.get(representation)()

    @property
    def level_trainable(self) -> bool:
        r"""
        Specifies whether the parameters weight and hidden are trainable or not.
        """
        return self._level_trainable

    @level_trainable.setter
    def level_trainable(self, val: bool) -> None:
        self._dual_param.requires_grad = val
        self._primal_param.requires_grad = val
        self._level_trainable = val

    @property
    def representation(self) -> str:
        return self._representation

    @property
    @abstractmethod
    def dim_feature(self) -> int:
        """
        Dimension of the explicit feature map if relevant.
        """
        pass

    @property
    def dim_output(self) -> int:
        r"""
        Output dimension.
        """
        if self._dim_output is None:
            raise utils.NotInitializedError(cls=self, message="The output dimension has not been initialized yet.")
        return self._dim_output

    @dim_output.setter
    def dim_output(self, val: int) -> None:
        self._dim_output = int(val)
        self._reset_primal()
        self._reset_dual()

    ##################################################################################################################
    ## HIDDEN

    @property
    def dual_param(self) -> Tensor:
        """
        Dual parameter of size [num_idx, dim_output].
        """
        if self._dual_param_exists:
            return self._dual_param.T[self.idx, :]
        raise utils.NotInitializedError(cls=self, message="No hidden values have been initialized or computed yet.")

    def update_dual(self, val: Tensor, idx_sample=None) -> None:
        # first verify the existence of the hidden values before updating them.
        with torch.no_grad():
            if not self._dual_param_exists:
                self._init_dual()

            if idx_sample is None:
                idx_sample = self.idx
            self._dual_param.copy_(val.data[idx_sample, :].T)
            if self._level_trainable and self._dual_param.grad is not None:
                self._dual_param.grad.zero_()
            self._reset_primal()

    @dual_param.setter
    def dual_param(self, val):
        # sets the parameter to an existing one
        if val is not None:
            with torch.no_grad():
                # to work on the stiefel manifold, the parameters are required to have the number of components as
                # first dimension
                val = utils.castf(val, dev=self._param_device)
                val = val.T
                if self._dual_param_exists and val.shape == self._dual_param.shape:
                    self._dual_param.copy_(val)
                    # zeroing the gradients if relevant
                    if self._level_trainable and self._dual_param.grad is not None:
                        self._dual_param.grad.zero_()
                else:
                    del self._dual_param
                    # torch.no_grad() does not affect the constructor
                    self._dual_param = torch.nn.Parameter(val, requires_grad=self._level_trainable)

                self._dim_output, self._num_h = self._dual_param.shape
            self._reset_primal()
        else:
            self._reset_dual()
            self._reset_primal()
            self._logger.info("The hidden value is unset.")

    @property
    def dual_trainable(self) -> bool:
        """
        Returns whether the hidden variables are trainable (a gradient can be computed on it).
        """
        return self._level_trainable

    @dual_trainable.setter
    def dual_trainable(self, val: bool):
        # changes the possibility of training the hidden values through backpropagation
        self._level_trainable = val
        self._dual_param.requires_grad = self._level_trainable

    @property
    def _dual_param_exists(self) -> bool:
        r"""
        Returns if this View has hidden variables attached to it.
        """
        return self._dual_param.nelement() != 0

    @property
    def attached(self) -> bool:
        r"""
        Boolean indicating whether the view is attached to another multi_view.
        """
        try:
            return self._attached_primal_param is not None
        except AttributeError:
            return False

    def attach_to(self, weight_fn) -> None:
        self._logger.debug(self.__repr__() + " is attached to a multi-view.")
        assert not self.attached, 'Cannot attach a view which is already attached'
        self._attached_primal_param = weight_fn

    def detach(self) -> None:
        self._logger.debug(self.__repr__() + ' is now detached.')
        w = self.primal_param
        self._attached = False
        self.primal_param = w

    ## WEIGHT
    @property
    def primal_param(self) -> Tensor:
        r"""
        Primal parameters of size [dim_feature, dim_output].
        """
        if self.attached:
            return self._attached_primal_param()
        else:
            if self._primal_param_exists:
                return self._primal_param.T
            raise utils.NotInitializedError(cls=self, message="No weight has been initialized or computed yet.")

    @primal_param.setter
    def primal_param(self, val):
        if self.attached:
            raise Exception("Cannot assign value of an attached view. Please change the value of the mother class "
                            "instead.")
        else:
            self._reset_dual()
            if val is not None:
                with torch.no_grad():
                    val = utils.castf(val, tensor=False, dev=self._primal_param.device)
                    if self._primal_param_exists and self._primal_param.shape == val.shape:
                        self._primal_param.copy_(val.T)
                        # zeroing the gradients if relevant
                        if self._level_trainable and self._primal_param.grad is not None:
                            self._primal_param.grad.data.zero_()
                    else:
                        del self._primal_param
                        # torch.no_grad() does not affect the constructor
                        self._primal_param = torch.nn.Parameter(val.T, requires_grad=self._level_trainable)

                    self._dim_output = self.primal_param.shape[1]
            else:
                self._reset_primal()
                self._logger.info("The weight is unset.")

    @property
    def _primal_param_exists(self) -> bool:
        try:
            return self._primal_param.nelement() != 0
        except AttributeError:
            return False

    def _update_primal_from_dual(self):
        raise NotImplementedError

    @abstractmethod
    def _update_dual_from_primal(self):
        raise NotImplementedError

    @property
    def _param_device(self) -> torch.device:
        return self._dual_param.device

    @abstractmethod
    def phi(self, x=None) -> Tensor:
        pass

    @abstractmethod
    def k(self, x=None) -> Tensor:
        pass

    def c(self, x=None) -> Tensor:
        phi = self.phi(x)
        return phi.T @ phi / self.num_idx

    def phiw(self, x=None, representation=None) -> Tensor:
        r"""
        :param x: Input, defaults to the sample (None).
        :type x: torch.Tensor [num, dim_input]
        :param representation: 'primal' or 'dual'.
        :type representation: str, optional
        :return: :math:`\phi(x)^\top W` or :math:`k(x)^top H`
        :rtype: torch.Tensor [num, dim_output]
        """
        def primal(x):
            return self.phi(x) @ self.W

        def dual(x):
            return self.kernel.k(x) @ self.H

        switcher = {"primal": primal,
                    "dual": dual}
        if representation in switcher:
            return switcher.get(representation)(x)
        else:
            raise utils.RepresentationError

    @property
    def Phi(self) -> Tensor:
        return self.phi()

    @property
    @abstractmethod
    def K(self) -> Tensor:
        pass

    @property
    def C(self) -> Tensor:
        return self.c()

    @property
    @abstractmethod
    def H(self) -> Tensor:
        pass

    @property
    @abstractmethod
    def W(self) -> Tensor:
        pass

    @property
    def PhiW(self) -> Tensor:
        return self.phiw()

    @property
    def dual_correlation(self) -> Tensor:
        r"""
        Correlation of the hidden variables :math:`\mathbf{h}^\top \mathbf{h}`. This should be the identity provided
        that the hidden variables lie on the Stiefel manifold.
        """
        return self.dual_param.T @ self.dual_param

    @property
    def primal_correlation(self) -> Tensor:
        r"""
        Correlation of the weights :math:`\mathbf{w}^\top \mathbf{w}`. This should be the identity provided
        that the weights lie on the Stiefel manifold.
        """
        return self.primal_param.T @ self.primal_param

    @property
    def dual_projector(self) -> Tensor:
        r"""
        Projector on the subspace spanned by the hidden variables :math:`\mathbf{h}\mathbf{h}^\top`.
        This is a rigorous projector provided its determinant is unity, e.g. when the hidden variables lie on the
        Stiefel manifold.
        """
        return self.dual_param @ self.dual_param.T

    @property
    def primal_projector(self) -> Tensor:
        r"""
        Projector on the subspace spanned by the weights :math:`\mathbf{w}\mathbf{w}^\top`.
        This is a rigorous projector provided its determinant is unity, e.g. when the weights lie on the Stiefel manifold.
        """
        return self.primal_param @ self.primal_param.T

    @abstractmethod
    def _forward(self, representation, x=None):
        pass

    def forward(self, x=None, representation=None):
        representation = utils.check_representation(representation, default=self._representation)
        name = f"forward_{id(x)}_{representation}"

        def get_level_key() -> str:
            if representation == self._representation:
                level_key = "forward_sample_default_representation" if x is None \
                    else "forward_oos_default_representation"
            else:
                level_key = "forward_sample_other_representation" if x is None \
                    else "forward_oos_other_representation"
            return level_key

        return self._get(name, level_key=lambda: get_level_key(), fun=lambda: self._forward(representation, x))



import torch
from torch.autograd import Variable
from torch.optim.lr_scheduler import ReduceLROnPlateau

from ..kernel._base_kernel import _BaseKernel as K
from ..feature.cache import Cache
from ._knn import knn
from ..utils import castf
from tqdm import tqdm


def iterative(obj, x0: torch.Tensor, num_iter: int = 50, lr=1.e-3, verbose: bool = False):
    r"""
    Minimizes to following problem for each point in order to find the preimage:

    .. math::
        \tilde{\mathbf{x}} = \mathrm{argmin}_{\mathbf{x}} \mathtt{obj}(\mathbf{x}).

    The method optimizes with an SGD algorithm.

    :param verbose: Shows the training loop. Defaults to ``False``.
    :type verbose: bool, optional
    :param obj: Objective to minimize.
    :param x0: Starting value for the optimization.
    :type x0: torch.Tensor [num_points, dim_input]
    :param num_iter: Number of iterations for the optimization process. Defaults to 50.
    :type num_iter: int, optional
    :param lr: Learning rate of the optimizer. Defaults to 0.001.
    :type lr: float, optional
    :return: Solution :math:`\tilde{\mathbf{x}}`
    :rtype: torch.Tensor [num_points, dim_input]
    """
    # PRELIMINARIES
    x0 = castf(x0)

    x = Variable(x0, requires_grad=True)
    optimizer = torch.optim.SGD([x], lr=lr)
    scheduler = ReduceLROnPlateau(optimizer, mode='min', factor=.8, patience=50, cooldown=50)

    # OPTIMIZE
    def closure():
        optimizer.zero_grad(set_to_none=True)
        loss = obj(x)
        loss.backward(retain_graph=False)
        return loss

    if verbose:
        epochs = tqdm(range(num_iter))
    else:
        epochs = range(num_iter)
    for _ in epochs:
        l = optimizer.step(closure)
        scheduler.step(l)
        last_lr = scheduler._last_lr[0]
        if verbose:
            epochs.set_description(f"Loss: {l:1.2e}, lr: {last_lr:1.1e}")
        if last_lr < 1.e-5:
            break

    # RETURN
    return x.data


def iterative_preimage_k(k_image: torch.Tensor, kernel: K, num_iter: int = 50, lr=1.e-3,
                         light_cache=True, verbose: bool = False) -> torch.Tensor:
    r"""
    Minimizes to following problem for each point in order to find the preimage:

    .. math::
        \tilde{\mathbf{x}} = \mathrm{argmin}_{\mathbf{x}} \big\lVert \mathtt{k\_image} - \mathtt{kernel.k(x)} \big\rVert_2^2

    The method optimizes with an SGD algorithm.

    :param verbose: Shows the training loop. Defaults to ``False``.
    :type verbose: bool, optional
    :param k_image: coefficients in the RKHS to be inverted.
    :type k_image: torch.Tensor [num_points, num_idx]
    :param kernel: kernel on which this RKHS is based.
    :type kernel: :py:class:`kerch.kernel.Kernel` instance.
    :param num_iter: Number of iterations for the optimization process. Defaults to 50.
    :type num_iter: int, optional
    :param lr: Learning rate of the optimizer. Defaults to 0.001.
    :type lr: float, optional
    :param light_cache: Specifies whether the cache has to made lighter during the pre-image to avoid keeping the
        statistics of each iteration. This results in a speedup. Defaults to ``True``.
    :type light_cache: bool, optional
    :return: Pre-image
    :rtype: torch.Tensor [num_points, dim_input]
    """
    k_image = castf(k_image).data

    # CHECK IF THE CACHE LEVEL HAS TO BE CHANGED
    cache_level = Cache._cache_level_switcher[kernel.cache_level]
    if cache_level > Cache._cache_level_switcher['light']:
        if light_cache:
            kernel.cache_level = 'light'
        else:
            kernel._logger.warning("The cache level is recommended to be at light at maximum in order to ease the "
                                   "memory load during the pre-image computation. It is temporarily being set to "
                                   "light. You can also set the argument light_cache to True to set it "
                                   "temporarily lower during the pre-image computation.")

    assert k_image.size(1) == kernel.num_idx, \
        f"Pre-image: the provided kernel coefficients ({k_image.size(1)}) do not correspond to the number " \
        f"of sample datapoints ({kernel.num_idx})."

    assert num_iter > 0, \
        f"The number of iterations num_iter ({num_iter}) must be strictly positive (num_iter > 0)."

    loss_fn = torch.nn.MSELoss()
    # x0 = torch.zeros(k_image.shape[0], kernel.dim_input, dtype=k_image.dtype)
    x0 = knn(dists=-k_image + k_image.max(), observations=kernel.current_sample)

    def obj(vals):
        k_current = kernel.k(x=vals)
        return loss_fn(k_current, k_image)

    sol = iterative(obj=obj, x0=x0, num_iter=num_iter, lr=lr, verbose=verbose)

    # SET BACK THE ORIGINAL CACHE LEVEL
    if (cache_level > Cache._cache_level_switcher['light']) and light_cache:
        kernel.cache_level = cache_level

    return sol


def iterative_preimage_phi(phi_image: torch.Tensor, kernel: K, num_iter: int = 50, lr=1.e-3,
                           light_cache=True, verbose: bool = False) -> torch.Tensor:
    r"""
    Minimizes to following problem for each point in order to find the preimage:

    .. math::
        \tilde{\mathbf{x}} = \mathrm{argmin}_{\mathbf{x}} \big\lVert \mathtt{phi\_image} - \mathtt{kernel.phi(x)} \big\rVert_2^2

    The method optimizes with an SGD algorithm.

    :param verbose: Shows the training loop. Defaults to ``False``.
    :type verbose: bool, optional
    :param phi_image: feature map image to be inverted.
    :type phi_image: torch.Tensor [num_points, dim_feature]
    :param kernel: kernel on which this RKHS is based.
    :type kernel: :py:class:`kerch.kernel.Kernel` instance.
    :param num_iter: Number of iterations for the optimization process. Defaults to 50.
    :type num_iter: int, optional
    :param lr: Learning rate of the optimizer. Defaults to 0.001.
    :type lr: float, optional
    :param light_cache: Specifies whether the cache has to made lighter during the pre-image to avoid keeping the
        statistics of each iteration. This results in a speedup. Defaults to ``True``.
    :type light_cache: bool, optional
    :return: Pre-image
    :rtype: torch.Tensor [num_points, dim_feature]
    """
    phi_image = castf(phi_image).data

    # CHECK IF THE CACHE LEVEL HAS TO BE CHANGED
    cache_level = Cache._cache_level_switcher[kernel.cache_level]
    if cache_level > Cache._cache_level_switcher['light']:
        if light_cache:
            kernel.cache_level = 'light'
        else:
            kernel._logger.warning("The cache level is recommended to be at light at maximum in order to ease the "
                                   "memory load during the pre-image computation. It is temporarily being set to "
                                   "light. You can also set the argument light_cache to True to set it "
                                   "temporarily lower during the pre-image computation.")

    assert phi_image.size(1) == kernel.num_idx, \
        f"Pre-image: the provided explicit feature map images dimensions ({phi_image.size(1)}) do not correspond to " \
        f"the feature dimension of the provided kernel ({kernel.dim_feature})."

    assert num_iter > 0, \
        f"The number of iterations num_iter ({num_iter}) must be strictly positive (num_iter > 0)."

    loss_fn = torch.nn.MSELoss()
    weights = phi_image @ kernel.Phi.T
    x0 = knn(dists=-weights, observations=kernel.current_sample)

    def obj(vals):
        phi_current = kernel.phi(x=vals)
        return loss_fn(phi_current, phi_image)

    sol = iterative(obj=obj, x0=x0, num_iter=num_iter, lr=lr, verbose=verbose)

    # SET BACK THE ORIGINAL CACHE LEVEL
    if (cache_level > Cache._cache_level_switcher['light']) and light_cache:
        kernel.cache_level = cache_level

    return sol



import torch
from torch import Tensor as T
from abc import ABCMeta, abstractmethod
from typing import Union

from ._level import _Level
from kerch import utils


class _KPCA(_Level):
    r"""
    Kernel Principal Component Analysis.
    """

    @utils.extend_docstring(_Level)
    @utils.kwargs_decorator({})
    def __init__(self, *args, **kwargs):
        super(_KPCA, self).__init__(*args, **kwargs)
        self._vals = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE),
                                        requires_grad=False)

    @property
    def vals(self) -> T:
        r"""
        Eigenvalues of the model. The model has to be fitted for these values to exist.
        """
        return self._vals

    @vals.setter
    def vals(self, val):
        val = utils.castf(val, tensor=False, dev=self._vals.device)
        self._vals.data = val

    def total_variance(self, as_tensor=False, normalize=True) -> Union[float, T]:
        r"""
        Total variance contained in the feature map. In primal formulation,
        this is given by :math:`\DeclareMathOperator{\tr}{tr}\tr(C)`, where :math:`C = \sum\phi(x)\phi(x)^\top` is
        the covariance matrix on the sample. In dual, this is given by :math:`\DeclareMathOperator{\tr}{tr}\tr(K)`,
        where :math:`K_{ij} = k(x_i,x_j)` is the kernel matrix on the sample.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional

        .. warning::
            For this value to strictly be interpreted as a variance, the corresponding kernel (or feature map)
            has to be normalized. In that case however, the total variance will amount to the dimension of the feature
            map in primal and the number of datapoints in dual.
        """
        if "_total_variance" not in self._cache:
            if self._representation == 'primal':
                self._cache["_total_variance"] = torch.trace(self.C)
            else:
                self._cache["_total_variance"] = torch.trace(self.K)
        var = self._cache["_total_variance"]
        if normalize:
            var /= self.num_idx
        if as_tensor:
            return var
        return var.detach().cpu().numpy()

    def model_variance(self, as_tensor=False, normalize=True) -> Union[float, T]:
        r"""
        Total variance learnt by the model given by the sum of the eigenvalues.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional

        .. warning::
            For this value to strictly be interpreted as a variance, the corresponding kernel (or feature map)
            has to be normalized. We refer to the remark of ``total_variance``.
        """
        var = torch.sum(self.vals)
        if normalize:
            var /= self.num_idx
        if as_tensor:
            return var
        return var.detach().cpu().numpy()

    def relative_variance(self, as_tensor=False) -> Union[float, T]:
        r"""
        Relative variance learnt by the model given by ```model_variance``/``relative_variance``.
        This number is always comprised between 0 and 1 and avoids any considerations on normalization.

        :param as_tensor: Indicated whether the variance has to be returned as a float or a torch.Tensor., defaults
            to ``False``
        :type as_tensor: bool, optional
        """
        var = self.model_variance(as_tensor=as_tensor, normalize=False) / \
              self.total_variance(as_tensor=as_tensor, normalize=False)
        return var

    ######################################################################################

    def _solve_primal(self) -> None:
        C = self.C

        if self.dim_output is None:
            self._dim_output = self.dim_feature
        elif self.dim_output > self.dim_feature:
            self._log.warning(f"In primal, the output dimension {self.dim_output} (the number of "
                              f"eigenvectors) must not exceed the feature dimension {self.dim_feature} (the dimension "
                              f"of the correlation matrix to be diagonalized). As this is the case here, the output "
                              f"dimension is reduced to {self.dim_feature}.")
            self.dim_output = self.dim_feature

        v, w = utils.eigs(C, k=self.dim_output, psd=True)

        self.weight = w
        self.vals = v

    def _solve_dual(self) -> None:
        K = self.K

        if self.dim_output is None:
            self._dim_output = self.num_idx
        elif self.dim_output > self.num_idx:
            self._log.warning(f"In dual, the output dimension {self.dim_output} (the number of "
                              f"eigenvectors) must not exceed the number of samples {self.num_idx} (the dimension "
                              f"of the kernel matrix to be diagonalized). As this is the case here, the output "
                              f"dimension is reduced to {self.num_idx}.")
            self.dim_output = self.num_idx

        v, h = utils.eigs(K, k=self.dim_output, psd=True)

        self.hidden = h
        self.vals = v

    @utils.extend_docstring(_Level.solve)
    def solve(self, sample=None, target=None, representation=None, **kwargs) -> None:
        r"""
        Solves the model by decomposing the kernel matrix or the covariance matrix in principal components
        (eigendecomposition).
        """
        # KPCA models don't require the target to be defined. This is verified.
        if target is not None:
            self._log.warning("The target value is discarded when fitting a KPCA model.")
        return _Level.solve(self,
                            sample=sample,
                            target=None,
                            representation=representation)

    ######################################################################################

    def _primal_obj(self, x=None) -> T:
        P = self.weight @ self.weight.T  # primal projector
        R = self._I_primal - P  # reconstruction
        C = self.c(x)  # covariance
        return torch.trace(R * C)  # reconstruction error on the covariance

    def _dual_obj(self, x=None) -> T:
        P = self.hidden @ self.hidden.T  # dual projector
        R = self._I_dual - P  # reconstruction
        K = self.k(x)  # kernel matrix
        return torch.trace(R * K)  # reconstruction error on the kernel

    ######################################################################################

    def _stiefel_parameters(self, recurse=True):
        # the stiefel optimizer requires the first dimension to be the number of eigenvectors
        super(_KPCA, self)._stiefel_parameters(recurse)
        if self._representation == 'primal':
            if self._weight_exists:
                yield self._weight
        else:
            if self._hidden_exists:
                yield self._hidden

    @utils.kwargs_decorator({"representation":None})
    def optimize(self, **kwargs) -> None:
        super(_KPCA, self).optimize(**kwargs)

        # once the optimization is done, the eigenvalues still have to be defined
        representation = utils.check_representation(kwargs["representation"], self._representation, cls=self)
        if representation == 'primal':
            self.vals = torch.diag(self.weight.T @ self.C @ self.weight)
        else:
            self.vals = torch.diag(self.hidden.T @ self.K @ self.hidden)

    @utils.kwargs_decorator({
        "representation": None
    })
    def fit(self, **kwargs):
        if not self.attached:
            if self.dim_output is None:
                representation = utils.check_representation(kwargs["representation"], self._representation, cls=self)
                if representation == "primal":
                    self._dim_output = self.dim_feature
                else:
                    self._dim_output = self.num_idx
            super(_KPCA, self).fit(**kwargs)

    ####################################################################################################################

    def loss(self, representation=None) -> T:
        r"""
        Reconstruction error on the sample.
        """
        representation = utils.check_representation(representation, self._representation, self)
        if representation == 'primal':
            U = self._weight    # transposed compared to weight
            M = self.C
        else:
            U = self._hidden    # transposed compared to hidden
            M = self.K
        loss = torch.trace(M) - torch.trace(U.T @ U @ M)
        return loss





# coding=utf-8
"""
Abstract class defining a general level in the toolbox.
"""
from __future__ import annotations

import torch
from typing import Iterator
from abc import ABCMeta, abstractmethod

from .logger import Logger
from .. import _GLOBALS
from ..utils import capitalize_only_first, extend_docstring


@extend_docstring(Logger)
class Module(Logger,
             torch.nn.Module,
             object,
             metaclass=ABCMeta):
    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    @abstractmethod
    def __init__(self, *args, **kwargs):
        # for some obscure reason, calling the super init does not lead to the call of both classes.
        # by consequence, we make the calls manually to each parents
        torch.nn.Module.__init__(self)
        Logger.__init__(self, *args, **kwargs)

    def __repr__(self):
        return capitalize_only_first(self.__str__())


    @property
    @extend_docstring(Logger.logging_level)
    def logging_level(self) -> int:
        return self._logger_internal.level

    @logging_level.setter
    def logging_level(self, level: int | None):
        if level is None:
            level = _GLOBALS["LOG_LEVEL"]
        self._logger_internal.setLevel(level)
        for child in self.children():
            if isinstance(child, Logger):
                child.logging_level = level

    def _euclidean_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        r"""
        Iterator yielding all parameters lying on the Euclidean manifold (standard optimization). The optimization is
        done onto :math:`\mathbb{R}^{n \times m}`, :math:`n` and :math:`m` depending on the size of each parameter.

        :param recurse: If ``True``, yields both the Euclidean parameters of this module and its potential children.
            otherwise, it only yields Euclidean parameters from this module. Defaults to ``True``.
        :type recurse: bool, optional
        :return: Euclidean parameters
        :rtype: Iterator[torch.nn.Parameter]
        """
        if recurse:
            for module in self.children():
                if isinstance(module, Module):
                    yield from module._euclidean_parameters(recurse)

    def _stiefel_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        r"""
        Iterator yielding all parameters that must lie on the Stiefel manifold (optimization is done onto that manifold).
        The Stiefel manifold corresponds to the orthonormal parameters :math:`U \in \mathrm{St}(n,m)`, i.e., all
        :math:`U \in \mathbb{R}^{n \times m}` such that :math:`U^\top U = I`. The dimensions :math:`n` and :math:`m` are
        proper to each parameter.

        :param recurse: If ``True``, yields both the Stiefel parameters of this module and its potential children.
            otherwise, it only yields Stiefel parameters from this module. Defaults to ``True``.
        :type recurse: bool, optional
        :return: Stiefel parameters
        :rtype: Iterator[torch.nn.Parameter]
        """
        if recurse:
            for module in self.children():
                if isinstance(module, Module):
                    yield from module._stiefel_parameters(recurse)

    def _slow_parameters(self, recurse=True) -> Iterator[torch.nn.Parameter]:
        r"""
        Iterator yielding all parameters lying on the Euclidean manifold (standard optimization). The optimization is
        done onto :math:`\mathbb{R}^{n \times m}`, :math:`n` and :math:`m` depending on the size of each parameter.

        The specificity of these slow Euclidean parameters is that they are better trained with a lower learning rate that the
        others, hence their name and the necessity to group them apart.

        :param recurse: If ``True``, yields both the slow (Euclidean) parameters of this module and its potential children.
            otherwise, it only yields slow (Euclidean) parameters from this module. Defaults to ``True``.
        :type recurse: bool, optional
        :return: Slow (Euclidean) parameters
        :rtype: Iterator[torch.nn.Parameter]
        """
        if recurse:
            for module in self.children():
                if isinstance(module, Module):
                    yield from module._slow_parameters(recurse)

    def manifold_parameters(self, recurse=True, type='euclidean') -> Iterator[torch.nn.Parameter]:
        r"""
        Iterator yielding the parameters of a specific type. A distinction is made between three types:

        * ``'euclidean'``:
            parameters lying on the Euclidean manifold (standard optimization). The optimization is
            done onto :math:`\mathbb{R}^{n \times m}`, :math:`n` and :math:`m` depending on the size of each parameter.
        * ``'stiefel'``:
            parameters that must lie on the Stiefel manifold (optimization is done onto that manifold).
            The Stiefel manifold corresponds to the orthonormal parameters :math:`U \in \mathrm{St}(n,m)`, i.e., all
            :math:`U \in \mathbb{R}^{n \times m}` such that :math:`U^\top U = I`. The dimensions :math:`n` and :math:`m` are
            proper to each parameter.
        * ``'slow'``:
            parameters lying on the Euclidean manifold (standard optimization). The optimization is
            done onto :math:`\mathbb{R}^{n \times m}`, :math:`n` and :math:`m` depending on the size of each parameter.
            The specificity of these slow Euclidean parameters is that they are better trained with a lower learning rate that the
            others, hence their name and the necessity to group them apart.

        We refer to the documentation of :doc:`../features/module` for further information.

        :param type: Which parameters group the method must return. The three values above are accepted. Defaults to ``'euclidean'``.
        :type type: str, optional
        :param recurse: If ``True``, yields both the specified parameters of this module and its potential children.
            otherwise, it only yields the specified parameters from this module. Defaults to ``True``.
        :type recurse: bool, optional
        :return: Parameters of the specified type
        :rtype: Iterator[torch.nn.Parameter]
        """
        switcher = {'euclidean': self._euclidean_parameters,
                    'stiefel': self._stiefel_parameters,
                    'slow': self._slow_parameters}
        gen = switcher.get(type, 'Invalid manifold name.')

        memo = set()
        for p in gen(recurse=recurse):
            if p not in memo:
                memo.add(p)
                yield p

    def before_step(self) -> None:
        r"""
        Specific operations to be performed before a training step. We refer to the documentation of
        :doc:`../features/module` for further information.
        """
        pass

    def after_step(self) -> None:
        r"""
            Specific operations to be performed after a training step. We refer to the documentation of
            :doc:`../features/module` for further information.
            """
        pass

    @property
    def hparams_variable(self) -> dict:
        r"""
        Variable hyperparameters of the module. By contrast with :py:attr:`hparams_fixed`, these are the values that are may change during
        the training and may be monitored at various stages during the training.
        If applicable, these can be kernel bandwidth parameters for example.

        .. note::

            All parameters that are potentially trainable, like a kernel bandwidth :math:`\sigma` for example, are
            included in this dictionary, even if the corresponding trainable argument is set to ``False``. In the
            latter case, they will be not evolve during training iterations, but will still be included in this
            dictionary.

        We refer to the documentation of :doc:`../features/module` for further information.
        """
        return {}

    @property
    def hparams_fixed(self) -> dict:
        r"""
        Fixed hyperparameters of the module. By contrast with :py:attr:`hparams_variable`, these are the values that are fixed and
        cannot possibly change during the training. If applicable, these can be specific architecture values for example.
        We refer to the documentation of :doc:`../features/module` for further information.
        """
        return {}



# coding=utf-8
import torch

from kerch.plot.matplotlib import *
from ..feature.sample import Sample
from ..kernel._base_kernel import _BaseKernel
from ..level._View import _View
from ..level._KPCA import _KPCA


def current_sample(sample: Sample, labels: bool = False) -> plt.Figure:
    assert isinstance(sample, Sample), 'The sample argument does not contain any sample.'
    return plot_sample(sample.current_sample, labels=labels)


def K(kernel, labels: bool = False) -> plt.Figure:
    assert isinstance(kernel, _BaseKernel), 'kernel must be a _BaseKernel object'
    return plot_K(kernel.K, labels=labels)


def C(kernel, labels: bool = False) -> plt.Figure:
    assert isinstance(kernel, _BaseKernel), 'kernel must be a _BaseKernel object'
    return plot_C(kernel.C, labels=labels)


def Phi(kernel, labels: bool = False) -> plt.Figure:
    assert isinstance(kernel, _BaseKernel), 'kernel must be a _BaseKernel object'
    return plot_Phi(kernel.Phi, labels=labels)


def W(view, labels: bool = False) -> plt.Figure:
    assert isinstance(view, _View), 'The argument view does not contain (primal) weights.'
    return plot_W(view.W, labels=labels)


def hidden(view, labels: bool = False) -> plt.Figure:
    assert isinstance(view, _View), 'The argument view does not contain hidden variables.'
    return plot_hidden(view.H, labels=labels)


def hidden_correlation(view, labels: bool = False) -> plt.Figure:
    assert isinstance(view, _View), 'The argument view does not contain hidden variables.'
    return plot_hidden_correlation(view.H, labels=labels)


def forward(view, labels: bool = False) -> plt.Figure:
    assert isinstance(view, _View), 'The argument view cannot perform a forward pass.'
    return plot_forward(view(), labels=labels)


def _get_coord(val) -> torch.Tensor:
    val_min, val_max = torch.min(val), torch.max(val)
    val_diff = val_max - val_min
    val_min, val_max = val_min - 0.1 * val_diff, val_max + 0.1 * val_diff
    return torch.linspace(val_min, val_max, 50, dtype=val.dtype, device=val.device)


def scatter(view, dims=None, dim_target: int = 0, heat_map: bool = False, labels: bool = False) -> plt.Figure:
    assert isinstance(view, _View), 'The view argument is incorrect.'

    if dims is None:
        dims = [0, 1]

    sample = view.current_sample.data.detach()
    x = sample[:, dims[0]]
    y = sample[:, dims[1]]

    cmap = plt.get_cmap('RdYlGn')
    fig = plt.figure()
    ax = fig.gca()

    # specify colors if the target exists
    try:
        t = view.target.data[:, dim_target].detach().cpu()
        ax.scatter(x.data.detach().cpu(), y.data.detach().cpu(), s=5, c=cmap(255 * (t == 1.)))
    except NotImplementedError or AttributeError:
        ax.scatter(x.data.detach().cpu(), y.data.detach().cpu(), s=5)

    # labels
    if labels:
        ax.set_xlabel(f"dim_output[{dims[0]}]")
        ax.set_ylabel(f"dim_output[{dims[1]}]")
        ax.set_title(f"Forward")

    # heat map
    if heat_map:
        if view.dim_input != 2:
            view._logger.warning('Cannot plot the heatmap for more than 2 input dimensions.')
            return fig
        x_coord = _get_coord(x)
        y_coord = _get_coord(y)
        x_grid, y_grid = torch.meshgrid(x_coord, y_coord)
        grid = torch.cat((x_grid.squeeze(), y_grid.squeeze()), dim=1)

        with torch.no_grad():
            eval_grid = view.forward(grid)[:, dim_target].data.detach().cpu()
        ax.contourf(x_grid.detach().cpu(), y_grid.detach().cpu(),
                    eval_grid.view(x_grid.shape), cmap=cmap, alpha=.5)

    return fig


def eigenvalues(kpca, labels: bool = False) -> plt.Figure:
    assert isinstance(kpca, _KPCA), 'The argument kpca does not contain eigenvalues.'
    return plot_eigenvalues(kpca.vals, labels=labels)



# coding=utf-8
from ._parser import get_args
from . import version, expe


def run():
    kwargs = get_args()
    action_name = kwargs.pop('action')[0]

    switcher = {'expe': expe}
    print(action_name)
    script = switcher.get(action_name)
    return script.run(**kwargs)


if __name__ == '__main__':
    run()



# coding=utf-8
"""
File containing the abstract kernel classes with transform.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021
"""
from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import List, Union

import torch
from torch import Tensor

from ..utils import kwargs_decorator, extend_docstring, castf, RepresentationError, equal
from ._base_kernel import _BaseKernel
from ..transform import TransformTree
from ..transform.all import MeanCentering, UnitSphereNormalization


@extend_docstring(_BaseKernel)
class Kernel(_BaseKernel, metaclass=ABCMeta):
    r"""
    :param kernel_transform: A list composed of the elements `'normalize'` or `'center'`. For example a centered
        cosine kernel which is centered and normalized in order to get a covariance matrix for example can be obtained
        by invoking a linear kernel with `default_transform = ['normalize', 'center', 'normalize']` or just a cosine
        kernel with `default_transform = ['center', 'normalize']`. Redundancy is automatically handled., defaults
        to `[]`.
    :type kernel_transform: List[str]
    """

    @abstractmethod
    @kwargs_decorator({
        "center": False,
        "normalize": False
    })
    def __init__(self, *args, **kwargs):
        super(Kernel, self).__init__(*args, **kwargs)

        kernel_transform = kwargs.pop('kernel_transform', [])

        # LEGACY SUPPORT
        if kwargs["center"]:
            self._logger.warning(
                "Argument center kept for legacy and will be removed in a later version. Please use the "
                "more versatile kernel_transform parameter instead.")
            kernel_transform.append("mean_centering")
        if kwargs["normalize"]:
            self._logger.warning(
                "Argument normalize kept for legacy and will be removed in a later version. Please use "
                "the more versatile kernel_transform parameter instead.")
            kernel_transform.append("unit_sphere_normalization")

        self._default_kernel_transform = self._simplify_transform(kernel_transform)

    @property
    def kernel_transform(self) -> TransformTree:
        r"""
        Default transform performed on the kernel
        """
        if self.explicit:
            return self._kernel_explicit_transform
        else:
            return self._kernel_implicit_transform

    @property
    def hparams_fixed(self) -> dict:
        return {'Default kernel transforms': self._default_kernel_transform, **super(Kernel, self).hparams_fixed}

    @property
    def _naturally_centered(self) -> bool:
        return False

    @property
    def _naturally_normalized(self) -> bool:
        return False

    @property
    def centered(self) -> bool:
        r"""
        Indicates whether the feature map is centered relative to its sample or equivalently is the kernel in centered
        in its RKHS space, spanned by the sample.
        """
        default = self.kernel_transform.final_transform
        return (self._naturally_centered and default == TransformTree) or default == MeanCentering

    @property
    def normalized(self) -> bool:
        r"""
        Indicates whether the feature map is centered relative to its sample or equivalently is the kernel in centered
        in its RKHS space, spanned by the sample.
        """
        default = self.kernel_transform.final_transform
        return (self._naturally_normalized and default == TransformTree) or default == UnitSphereNormalization

    @property
    def _required_transform(self) -> Union[List, None]:
        return None

    def _simplify_transform(self, transform=None) -> List:
        if transform is None:
            transform = []

        # add requirements
        if self._required_transform is not None:
            transform.append(self._required_transform)

        # remove same following elements
        transform = TransformTree.beautify_transform(transform)

        # remove unnecessary operation if kernel does it by default
        try:
            if self._naturally_normalized and transform[0] == UnitSphereNormalization:
                transform.pop(0)
        except IndexError:
            pass

        return transform

    def _get_transform(self, transform=None) -> List:
        if transform is None:
            return self._default_kernel_transform
        return self._simplify_transform(transform)

    @property
    def _kernel_explicit_transform(self) -> TransformTree:
        def fun():
            return TransformTree(explicit=True,
                                 sample=self._explicit_with_none,
                                 default_transform=self._default_kernel_transform,
                                 cache_level=self._cache_level)

        return self._get("kernel_explicit_transform", level_key="kernel_explicit_transform", fun=fun)

    @property
    def _kernel_implicit_transform(self) -> TransformTree:
        def fun():
            return TransformTree(explicit=False,
                                 sample=self._implicit_with_none,
                                 default_transform=self._default_kernel_transform,
                                 cache_level=self._cache_level,
                                 implicit_self=self._implicit_self)

        return self._get("kernel_implicit_transform", level_key="kernel_implicit_transform", fun=fun)

    def _phi(self):
        r"""
        Returns the explicit feature map with default centering and normalization. If already computed, it is
        recovered from the cache.
        """

        def fun():
            self._check_sample()
            return self._kernel_explicit_transform.projected_sample

        return self._get("phi", level_key="sample_phi", fun=fun)

    def _C(self) -> Tensor:
        r"""
        Returns the covariance matrix with default centering and normalization. If already computed, it is recovered
        from the cache.
                """

        def fun():
            scale = 1 / self.num_idx
            phi = self._phi()
            return scale * phi.T @ phi

        return self._get("C", level_key="sample_C", fun=fun)

    def _K(self, explicit=None, overwrite: bool = False) -> Tensor:
        r"""
        Returns the kernel matrix with default centering and normalization. If already computed, it is recovered from
        the cache.

        :param explicit: Specifies whether the explicit or implicit formulation has to be used. Always uses
            the explicit if available.
        :param overwrite: By default, the kernel matrix is recovered from cache if already computed. Force overwrites
            this if True., defaults to False
        """

        def fun(explicit):
            self._check_sample()
            if explicit is None: explicit = self.explicit
            if explicit:
                phi = self._phi()
                return phi @ phi.T
            else:
                return self._kernel_implicit_transform.projected_sample

        return self._get("K", level_key="sample_K", fun=lambda: fun(explicit), overwrite=overwrite)

    # ACCESSIBLE METHODS
    def phi(self, x=None, transform=None) -> Tensor:
        r"""
        Returns the explicit feature map :math:`\phi(x)` of the specified points.

        :param x: The datapoints serving as input of the explicit feature map. If `None`, the sample will be used.
            Defaults to ``None``.
        :type x: Tensor[num_x, dim_input], optional
        :return: Explicit feature map :math:`\phi(x)` of the specified points.
        :rtype: Tensor[num_x, dim_feature]
        :raises: ExplicitError
        """
        x = castf(x)
        transform = self._get_transform(transform)
        return self._kernel_explicit_transform.apply(x=self.transform_input(x), transform=transform)

    def k(self, x=None, y=None, explicit=None, transform=None) -> Tensor:
        r"""
        Returns a kernel matrix, either of the sample, either out-of-sample, either fully out-of-sample.

        .. math::
            K = [k(x_i,y_j)]_{i,j=1}^{\texttt{num_x}, \texttt{num_y}},

        with :math:`\{x_i\}_{i=1}^\texttt{num_x}` the out-of-sample points (`x`) and :math:`\{y_i\}_{j=1}^\texttt{num_y}` the sample points
        (`y`).

        .. note::
            In the case of centered kernels on an out-of-sample, this computation is more expensive as it requires to center according to
            the sample data, which implies computing a statistic on the out-of-sample kernel matrix and thus
            also computing it.

        :param x: Out-of-sample points (first dimension). If `None`, the default sample will be used. Defaults to ``None``
        :param y: Out-of-sample points (second dimension). If `None`, the default sample will be used. Defaults to ``None``

        :type x: Tensor[num_x, dim_input], optional
        :type y: Tensor[num_y, dim_input], optional

        :return: Kernel matrix
        :rtype: Tensor[num_x, num_y]

        :raises: ExplicitError
        """
        # if x is None and y is None:
        #     return self.K
        # in order to get the values in the correct format (e.g. coming from numpy)
        if explicit is None:
            explicit = self.explicit

        x = castf(x)
        y = castf(y)
        transform = self._get_transform(transform)
        if explicit:
            phi_x = self._kernel_explicit_transform.apply(x=self.transform_input(x),
                                                          transform=transform)
            if equal(x, y):
                phi_y = phi_x
            else:
                phi_y = self._kernel_explicit_transform.apply(x=self.transform_input(y),
                                                              transform=transform)
            return phi_x @ phi_y.T
        else:  # implicit
            if equal(x, y):
                x = self.transform_input(x)
                return self._kernel_implicit_transform.apply(x=x,
                                                             y=x,
                                                             transform=transform)
            else:
                return self._kernel_implicit_transform.apply(x=self.transform_input(x),
                                                             y=self.transform_input(y),
                                                             transform=transform)

    def c(self, x=None, transform=None) -> Tensor:
        r"""
        Out-of-sample explicit matrix.

        .. math::
            C = \frac{1}{\texttt{num_x}}\sum_{i}^{\texttt{num_x}} \phi(x_i)\phi(x_i)^\top.

        :param x: Out-of-sample points (first dimension). If `None`, the default sample will be used., defaults to `None`
        :type x: Tensor(num_x,dim_input), optional

        :return: Explicit matrix
        :rtype: Tensor[dim_feature, dim_feature]
        """

        phi = self.phi(x=x, transform=transform)
        scale = 1 / phi.shape[0]
        return scale * phi.T @ phi

    def forward(self, x, representation="dual") -> Tensor:
        """
        Passes datapoints through the kernel.

        :param x: Datapoints to be passed through the kernel.
        :param representation: Chosen representation. If `dual`, an out-of-sample kernel matrix is returned. If
            `primal` is specified, it returns the explicit feature map., defaults to `dual`

        :type x: Tensor(,dim_input)
        :type representation: str, optional

        :return: Out-of-sample kernel matrix or explicit feature map depending on `representation`.

        :raises: RepresentationError
        """

        def primal(x):
            return self.phi(x)

        def dual(x):
            return self.k(x)

        switcher = {"primal": primal,
                    "dual": dual}

        fun = switcher.get(representation, RepresentationError)
        return fun(x)

    def cov(self, x=None) -> Tensor:
        r"""
        Returns the covariance matrix fo the provided input.

        :param x: Out-of-sample points (first dimension). If `None`, the default sample will be used.
            Defaults to `None`.
        :type x: Tensor[N, dim_input], optional

        :return: Covariance matrix
        :rtype: Tensor[dim_feature, dim_feature]
        """
        transform = self._default_kernel_transform.copy()
        try:
            if not (self._default_kernel_transform[-1] == 'center'):
                transform.append('center')
        except IndexError:
            transform.append('center')

        transform = self._simplify_transform(transform)
        return self.c(x=x, transform=transform)

    def corr(self, x=None) -> Tensor:
        """
        Returns the correlation matrix fo the provided input.

        :param x: Out-of-sample points (first dimension). If `None`, the default sample will be used., defaults to `None`
        :type x: Tensor[N, dim_input], optional

        :return: Correlation matrix
        :rtype: Tensor[dim_feature, dim_feature]
        """
        cov = self.cov(x=x)
        var = torch.sqrt(torch.diag(cov))[:, None]
        return cov / (var * var.T)

    def explicit_preimage(self, phi_image: torch.Tensor | None = None, method: str = 'explicit',
                          **kwargs) -> torch.Tensor:
        r"""
        Computes a pre-image of an explicit feature map of the kernel, given by ``phi_image``.
        Different methods are available:

        * ``'explicit'``: Uses an explicit implementation specific to the kernel (if available). This is always preferable if available.
        * ``'knn'``: Nearest neighbors. We refer to :py:func:`kerch.method.knn` for more details.
        * ``'smoother'``: Kernel smoothing. We refer to :py:func:`kerch.method.smoother` for more details
        * ``'iterative'``: Iterative optimization. We refer to :py:func:`kerch.method.iterative_preimage_phi` for more details

        :param phi_image: Explicit feature map image to be inverted. If not specified (``None``), the explicit feature map on the sample is used.
        :type phi_image: torch.Tensor [num_points, dim_feature], optional
        :param method: Pre-image method to be used. Defaults to ``'explicit'``.
        :type method: str, optional
        :param \**kwargs: Additional parameters of the pre-image method used. Please refer to its documentation for
            further details.
        :type \**kwargs: dict, optional
        :return: Pre-image
        :rtype: torch.Tensor [num_points, dim_input]

        """
        phi_image = castf(phi_image)
        if phi_image is None:
            phi_image = self.phi()

        method = method.lower()
        if method == 'explicit':
            phi = self._kernel_explicit_transform.revert(phi_image)
            x_tilde = self._explicit_preimage(phi)
            return self.sample_transform.revert(x_tilde)
        elif method == 'knn':
            from ..method import knn
            k_image = phi_image @ self.Phi.T
            return knn(dists=-k_image, observations=self.current_sample, **kwargs)
        elif method == 'smoother':
            from ..method import smoother
            k_image = phi_image @ self.Phi.T
            return smoother(coefficients=k_image, observations=self.current_sample, **kwargs)
        elif method == 'iterative':
            from ..method import iterative_preimage_phi
            return iterative_preimage_phi(phi_image=phi_image, kernel=self, **kwargs)
        else:
            raise AttributeError('Unknown or non-implemented preimage method.')



# coding=utf-8
DEFAULT_KERNEL_TYPE = 'rbf'
DEFAULT_CACHE_LEVEL = {"forward_sample_default_representation": "light",
                       "forward_sample_other_representation": "normal",
                       "forward_oos_default_representation": "heavy",
                       "forward_oos_other_representation": "total",
                       "sample_phi": "light",
                       "sample_C": "light",
                       "sample_K": "light",
                       "Level_I_default_representation": "normal",
                       "Level_I_other_representation": "total",
                       "PPCA_B_primal": "normal",
                       "PPCA_B_dual": "normal",
                       "PPCA_Inv_primal": "normal",
                       "PPCA_Inv_dual": "normal",
                       "KPCA_total_variance_default_representation": "normal",
                       "KPCA_total_variance_other_representation": "total",
                       "Level_subloss_default_representation": "normal",
                       "Level_subloss_other_representation": "total",
                       "sample_transform": "none",
                       "kernel_explicit_transform": "none",
                       "kernel_implicit_transform": "none",
                       "Wasserstein_kernel_dist": "normal",
                       "transform_sample_data_default": "light",
                       "transform_sample_data_nondefault": "normal",
                       "transform_sample_statistics_default": "light",
                       "transform_sample_statistics_nondefault": "light",
                       "transform_oos_data_level_default": "heavy",
                       "transform_oos_data_level_nondefault": "total",
                       "transform_oos_statistics_default": "normal",
                       "transform_oos_statistics_nondefault": "heavy",
                       "_poly_explicit_fun": "light",
                       "_rsf_piv": "normal",
                       "_nystrom_elements": "light"
                       }



# coding=utf-8
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from datetime import datetime
import socket
import os
import torch

from ..opt import Optimizer
from ..model.Model import Model


class Watcher(metaclass=ABCMeta):
    r"""
    :param model: Model to watch (monitor, log and save)
    :param opt: Optimizer used in the training
    :param expe_name: Name of the experiment
    :param verbose: Verbosity boolean. Defaults to False.
    :param num_epochs_loss: The losses will be logged every `num_epochs_loss`. Defaults to 1.
    :param num_epochs_save: The model will be saved every `num_epochs_save`. Defaults to 1.
    :param num_epochs_params: The model parameters will be logged every `num_epochs_params`. Defaults to 1.
    :param num_epochs_plot: The watched properties will be logged/plotted every `num_epochs_plot`. Defaults to 1.
    :type model: kerch.model.Model
    :type opt: kerch.opt.Optimizer
    :type expe_name: str
    :type verbose: bool, optional
    :type num_epochs_loss: int, optional
    :type num_epochs_save: int, optional
    :type num_epochs_params: int, optional
    :type num_epochs_plot: int, optional
    """

    def __init__(self, model: Model, opt: Optimizer | None, expe_name: str, verbose: bool = False, **kwargs):
        super(Watcher, self).__init__(**kwargs)
        self._verbose = verbose
        self._num_epochs_loss = kwargs.pop('num_epochs_loss', 1)
        self._num_epochs_params = kwargs.pop('num_epochs_params', 1)
        self._num_epochs_save = kwargs.pop('num_epochs_save', 1)
        self._num_epochs_plot = kwargs.pop('num_epochs_plot', 1)

        # MODEL
        self._model = model
        assert isinstance(model, Model), 'The model argument is not a correct Kerch Model.'

        # DIRECTORY
        current_time = datetime.now().strftime('%b%d_%H-%M-%S')
        self._id = current_time + '_' + socket.gethostname()
        assert isinstance(expe_name, str), 'The expe_name argument must be a string.'
        self._expe_name = expe_name
        self._expe_id = self._id
        self._dir_project = os.path.join("kerch", self._expe_name, self._expe_id)
        self._dir_model = os.path.join(self._dir_project, "model")

        # OPTIMIZER
        if opt is not None:
            assert isinstance(opt, Optimizer), 'The opt argument is not a correct Kerch Optimizer.'
        self._opt = opt

    @property
    def verbose(self) -> bool:
        return self._verbose

    @property
    def model(self) -> Model:
        r"""
        Model used for logging purposes
        """
        return self._model

    @property
    def opt(self) -> Optimizer:
        r"""
        Optimizer used for logging purposes
        """
        return self._opt

    @property
    def expe_name(self) -> str:
        r"""
        Experiment name
        """
        return self._expe_name

    @property
    def expe_id(self) -> str:
        r"""
        Unique experiment identifier
        """
        return self._expe_id

    @property
    def dir_project(self) -> str:
        r"""
        Relative path of the project directory.
        """
        return self._dir_project

    def save_model(self, epoch: int) -> str:
        r"""
        Saves the model current state.

        :param epoch: Current epoch number. If None, it is considered to be the final state. Defaults to None.
        :type epoch: int, optional
        :return: The relative path where the model is saved
        :rtype: str
        """
        if not os.path.exists(self._dir_model):
            os.makedirs(self._dir_model)

        save_name = 'epoch-' + str(epoch)
        save_path = os.path.join(self._dir_model, save_name + '.pth')
        torch.save(self._model.state_dict(), save_path)
        return save_path

    def finish(self) -> str:
        r"""
        Cleans all the checkpoints of the model states, but the final one.

        :return: The relative path where the final model has been saved.
        :rtype: str
        """
        filelist = os.listdir(self._dir_model)
        last_epoch = max((int(file[len('epoch-'):-len('.pth')]) for file in filelist))
        last_file = 'epoch-' + str(last_epoch) + '.pth'
        filelist = {file for file in filelist if file != last_file}
        for f in filelist:
            os.remove(os.path.join(self._dir_model, f))
        return os.path.relpath(os.path.join(self._dir_model, last_file))

    @abstractmethod
    def update(self,
               epoch: int,
               objective_loss: float,
               training_error: float | None = None,
               validation_error: float | None = None,
               test_error: float | None = None
               ) -> None:
        pass



import matplotlib.pyplot as plt

from ..rkm._kpca import _KPCA
from ..rkm._view import _View

def eigvals(kpca:_KPCA, log=True, ax=None) -> None:
    vals = kpca.vals.detach().cpu()

    indep = ax is None
    if indep:
        plt.figure()
        ax = plt.gca()

    ax.bar(range(len(vals)),vals)
    ax.set_title('Eigenvalues of ' + kpca.__str__())
    ax.set_xlabel('Different components')
    ax.set_ylabel('Eigenvalue magnitude')

    if log:
        ax.set_yscale('log')

    if indep:
        plt.show()

def correlation_hidden(view:_View, ax=None) -> None:
    h = view.hidden
    corr = (h.T @ h).detach().cpu()

    indep = ax is None
    if indep:
        plt.figure()
        ax = plt.gca()

    im = ax.imshow(corr)
    ax.set_title('Correlation of the hidden variables')
    plt.colorbar(im)

    if indep:
        plt.show()




import numpy as np

from ..model.Model import Model
from .._logger import _GLOBAL_LOGGER
from ..rkm._level import _Level
from matplotlib import pyplot as plt

def classifier_level(lvl: _Level, dims=None):
    if dims is None:
        dims = [0, 1]

    # get the sample
    try:
        sample = lvl.sample.detach().cpu().numpy()
        x = sample[:,dims[0]]
        y = sample[:,dims[1]]
        t = lvl.targets.detach().cpu().numpy()
    except AttributeError:
        _GLOBAL_LOGGER._log.warning("Could not plot based on the training points as these seem not "
                                    "to exist.")
        return

    # create grid
    def _get_coord(val):
        val_min, val_max = np.min(val), np.max(val)
        val_diff = val_max - val_min
        val_min, val_max = val_min - 0.1 * val_diff, val_max + 0.1 * val_diff
        return np.linspace(val_min, val_max, 50)

    x_coord = _get_coord(x)
    y_coord = _get_coord(y)
    x_grid, y_grid = np.meshgrid(x_coord, y_coord)
    grid = np.concatenate((np.expand_dims(x_grid.flatten(), axis=1),
                           np.expand_dims(y_grid.flatten(), axis=1)),
                          axis=1)

    # evaluate
    eval_grid = lvl.forward(grid).detach().cpu()

    # PLOTS
    cmap = plt.get_cmap('RdYlGn')
    plt.figure()
    plt.scatter(x, y, s=5, c=cmap(255*(t==1.)))
    plt.contourf(x_grid, y_grid, eval_grid.view(x_grid.shape), cmap=cmap, alpha=.5)
    plt.xlabel('Dimension 1')
    plt.ylabel('Dimension 2')
    plt.title(lvl.__class__.__name__)
    plt.show()

def model(mdl:Model, dims=None):
    if dims is None:
        dims = [0, 1]

    # get the sample
    try:
        sample = mdl._training_data.cpu().numpy()
        x = sample[:,dims[0]]
        y = sample[:,dims[1]]
        t = mdl._training_labels
    except AttributeError:
        _GLOBAL_LOGGER._log.warning("Could not plot based on the training points as these seem not "
                                    "to exist.")
        return

    # create grid
    def _get_coord(val):
        val_min, val_max = np.min(val), np.max(val)
        val_diff = val_max - val_min
        val_min, val_max = val_min - 0.1 * val_diff, val_max + 0.1 * val_diff
        return np.linspace(val_min, val_max, 50)

    x_coord = _get_coord(x)
    y_coord = _get_coord(y)
    x_grid, y_grid = np.meshgrid(x_coord, y_coord)
    grid = np.concatenate((np.expand_dims(x_grid.flatten(), axis=1),
                           np.expand_dims(y_grid.flatten(), axis=1)),
                          axis=1)

    # evaluate
    eval_grid = mdl.forward(grid)

    # PLOTS
    cmap = plt.get_cmap('RdYlGn')
    plt.figure()
    plt.scatter(x, y, s=5, c=cmap(255*(t==1.)))
    plt.contourf(x_grid, y_grid, eval_grid.view(x_grid.shape), cmap=cmap, alpha=.5)
    plt.xlabel('Dimension 1')
    plt.ylabel('Dimension 2')
    plt.title(mdl.__class__.__name__)
    plt.show()



# coding=utf-8
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from random import shuffle
from torch.utils.data import TensorDataset

from ..utils import kwargs_decorator, castf


class _LearningSet(metaclass=ABCMeta):
    @kwargs_decorator({'shuffle': True})
    def __init__(self, name: str, dim_data: int, dim_labels: int, range:int = None, **kwargs):
        super(_LearningSet, self).__init__()
        self._name = name
        self._dim_data = dim_data
        self._dim_labels = dim_labels
        self._plot_range: list[int] | None = range
        self._shuffle: bool = kwargs.pop('shuffle', True)

    def __str__(self):
        return self._name + \
            f" (Training: {len(self.training_set)}, Validation: {len(self.validation_set)}, Test: {len(self.test_set)})"

    def _init_datasets(self,
                       training_data, training_labels,
                       validation_data, validation_labels,
                       test_data, test_labels):
        assert training_data.shape[0] == training_labels.shape[0], \
            'The number of training datapoints and labels is inconsistent.'
        self._training_set = TensorDataset(castf(training_data), castf(training_labels))

        assert validation_data.shape[0] == validation_labels.shape[0], \
            'The number of validation datapoints and labels is inconsistent.'
        self._validation_set = TensorDataset(castf(validation_data), castf(validation_labels))

        assert test_data.shape[0] == test_labels.shape[0], \
            'The number of test datapoints and labels is inconsistent.'
        self._test_set = TensorDataset(castf(test_data), castf(test_labels))

    @property
    def info(self) -> dict:
        return {'name': self._name,
                'range': self._plot_range,
                'dim_data': self._dim_data,
                'dim_labels': self._dim_labels}

    @property
    def training_set(self) -> TensorDataset | None:
        return self._training_set

    @property
    def validation_set(self) -> TensorDataset | None:
        return self._validation_set

    @property
    def test_set(self) -> TensorDataset | None:
        return self._test_set

    @property
    def validation_exists(self) -> bool:
        return len(self._validation_set) > 0

    @property
    def test_exists(self) -> bool:
        return len(self._test_set) > 0

    def _indices(self, sizes: list[int], total=None) -> list[list]:
        r"""
        Splits the value in multiple random datasets based on props.
        """
        if total is None:
            total = sum(sizes)
        idx = list(range(total))
        if self._shuffle:
            shuffle(idx)

        idx_list = []
        pos_start = 0
        for size in sizes:
            if size == 0.:
                idx_list.append([])
            else:
                pos_end = pos_start + size
                idx_list.append(idx[pos_start:pos_end])
                pos_start = pos_end
        return idx_list


class _LearningSetTrain(_LearningSet, metaclass=ABCMeta):
    @kwargs_decorator({'num_training': 100,
                       'num_validation': 20,
                       'num_test': 20})
    def __init__(self, *args, **kwargs):
        super(_LearningSetTrain, self).__init__(*args, **kwargs)
        sizes = [kwargs['num_training'],
                 kwargs['num_validation'],
                 kwargs['num_test']]
        num = sum(sizes)

        data, labels = self._training(num)
        idx_list = self._indices(sizes, total=data.shape[0])
        idx_training = idx_list[0]
        idx_validation = idx_list[1]
        idx_test = idx_list[2]

        self._init_datasets(training_data=data[idx_training],
                            training_labels=labels[idx_training],
                            validation_data=data[idx_validation],
                            validation_labels=labels[idx_validation],
                            test_data=data[idx_test],
                            test_labels=labels[idx_test])

    @abstractmethod
    def _training(self, num):
        pass


class _LearningSetTrainTest(_LearningSet, metaclass=ABCMeta):
    @kwargs_decorator({'num_training': 100,
                       'num_validation': 20,
                       'num_test': 20})
    def __init__(self, *args, **kwargs):
        super(_LearningSetTrainTest, self).__init__(*args, **kwargs)
        sizes = [kwargs['num_training'],
                 kwargs['num_validation']]
        num = sum(sizes)

        training_data, training_labels = self._training(num)
        test_data, test_labels = self._test(kwargs['num_test'])
        idx_list = self._indices(sizes, total=training_data.shape[0])
        idx_training = idx_list[0]
        idx_validation = idx_list[1]
        idx_test = self._indices(kwargs['num_test'], total=test_data.shape[0])

        self._init_datasets(training_data=training_data[idx_training],
                            training_labels=training_labels[idx_training],
                            validation_data=training_data[idx_validation],
                            validation_labels=training_labels[idx_validation],
                            test_data=test_data[idx_test],
                            test_labels=test_labels[idx_test])

    @abstractmethod
    def _training(self, num):
        pass

    @abstractmethod
    def _test(self, num):
        pass



# coding=utf-8
from __future__ import annotations

from abc import ABCMeta, abstractmethod
import os

from .Watcher import Watcher

class Plotter(Watcher, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super(Plotter, self).__init__(*args, **kwargs)
        if self._plotter_name is None:
            self._dir_plotter = self.dir_project
        else:
            self._dir_plotter = os.path.join(self._dir_project, self._plotter_name)

    @property
    @abstractmethod
    def _plotter_name(self) -> str | None:
        pass

    @property
    def dir_plotter(self) -> str:
        r"""
        Relative path of the plotter directory.
        """
        return self._dir_plotter


from .lssvm import LSSVM
from .. import utils
import torch


class Ridge(LSSVM):
    @utils.extend_docstring(LSSVM)
    @utils.kwargs_decorator({
        "bias": False
    })
    def __init__(self, **kwargs):
        if kwargs["bias"]:
            kwargs["bias"] = False
            self._log.warning('A ridge regression has no bias term. '
                              'The bias parameter is overwritten to False.')
        super(Ridge, self).__init__(**kwargs)

    def _solve_primal(self, target=None) -> None:
        C = self.kernel.C
        phi = self.kernel.phi_sample
        dev = C.device
        dim_output = phi.shape[1]

        I = torch.eye(dim_output,
                      dtype=utils.FTYPE,
                      device=dev)
        Y = phi.t() @ target

        A = C + (1 / self._gamma) * I

        sol = torch.linalg.solve(A, Y)
        self.weight = sol

    def _solve_dual(self, target=None) -> None:
        K = self.kernel.K
        dev = K.device

        I = torch.eye(self.num_sample,
                      dtype=utils.FTYPE,
                      device=dev)

        A = K + (1 / self._gamma) * I

        sol = torch.linalg.solve(A, target)
        self.hidden = sol



# coding=utf-8
from .decorators import (kwargs_decorator as kwargs_decorator, extend_docstring as extend_docstring)
from .cast import (castf as castf, casti as casti, check_representation as check_representation,
                   capitalize_only_first as capitalize_only_first)
from .type import (set_eps as set_eps, set_ftype as set_ftype, set_itype as set_itype, gpu_available as gpu_available,
                   FTYPE as FTYPE, ITYPE as ITYPE, EPS as EPS)
from .math import eigs as eigs
from .errors import (ImplicitError as ImplicitError,
                     ExplicitError as ExplicitError,
                     RepresentationError as RepresentationError,
                     BijectionError as BijectionError,
                     NotInitializedError as NotInitializedError,
                     MultiViewError as MultiViewError,
                     KerchError as KerchError)
from .tensor import (eye_like as eye_like, ones_like as ones_like, equal as equal)
from .defaults import (DEFAULT_KERNEL_TYPE as DEFAULT_KERNEL_TYPE,
                       DEFAULT_CACHE_LEVEL as DEFAULT_CACHE_LEVEL)
from .dict import reverse_dict as reverse_dict


import torch
from torch import Tensor as T
from abc import ABCMeta, abstractmethod

import kerch.opt
from ._view import _View
from kerch import utils


class _Level(_View, metaclass=ABCMeta):
    r"""
    :param eta: :math:`\eta`., defaults to 1.
    :param representation: Chosen representation, "primal" or "dual"., defaults to "dual".

    :type eta: double, optional
    :type representation: str, optional
    """

    @utils.extend_docstring(_View)
    @utils.kwargs_decorator({
        "eta": 1.,
    })
    def __init__(self, *args, **kwargs):
        super(_Level, self).__init__(*args, **kwargs)
        self.eta = kwargs["eta"]

    @property
    def _I_primal(self) -> T:
        r"""
            To avoid multiple reinitializations, certainly when working on GPU, the value is stored in memory the first
            time it is called, to be re-used later."
        """
        if "I_primal" not in self._cache:
            self._cache["I_primal"] = torch.eye(self.dim_feature, dtype=utils.FTYPE)
        return self._cache["I_primal"]

    @property
    def _I_dual(self) -> T:
        r"""
                To avoid multiple reinitializations, certainly when working on GPU, the value is stored in memory the first
                time it is called, to be re-used later."
        """
        if "I_dual" not in self._cache:
            self._cache["I_dual"] = torch.eye(self.num_idx, dtype=utils.FTYPE)
        return self._cache["I_dual"]

    ####################################################################################################################

    @abstractmethod
    def _solve_primal(self) -> None:
        r"""
        Solves the dual formulation on the sample.
        """
        pass

    @abstractmethod
    def _solve_dual(self) -> None:
        r"""
        Solves the primal formulation on the sample.
        """
        pass

    def solve(self, sample=None, targets=None, representation=None, **kwargs) -> None:
        r"""
        Fits the model according to the input ``sample`` and output ``target``. Many models have both a primal and
        a dual formulation to be fitted.

        :param representation: Representation of the model (``"primal"`` or ``"dual"``)., defaults to ``"dual"``.
        :type representation: str, optional
        """

        # self._log.debug("The fitting is always done on the full sample dataset, regardless of the stochastic state.")
        # set the sample to input (always works for the underlying kernel)

        if sample is not None:
            self.init_sample(sample)
        if targets is not None:
            self.targets = targets

        # verify that the sample has been initialized
        if self._num_total is None:
            self._log.error("Cannot perform fitting as no input has been provided nor a sample already exists")
            return

        # check the representation is correct and set it to the default Level value if None
        representation = utils.check_representation(representation, default=self._representation, cls=self)

        # execute the corresponding fitting
        switcher = {"primal": self._solve_primal,
                    "dual": self._solve_dual}
        fun = switcher.get(representation)
        return fun()

    ####################################################################################################################

    @utils.kwargs_decorator({
        "method": "exact",
    })
    def fit(self, **kwargs):
        switcher = {"exact": self.solve,
                    "optimize": self.optimize}
        switcher.get(kwargs["method"], 'Invalid method name (must be solve or optimize')(**kwargs)

    @utils.kwargs_decorator({
        "representation": None,
        # "loss": "classic",
        "maxiter": 10000,
        "verbose": True
    })
    def optimize(self, **kwargs) -> None:
        representation = utils.check_representation(kwargs["representation"], self._representation, cls=self)

        # PRELIMINARIES
        self.init_parameters(representation, overwrite=False)
        opt = kerch.opt.Optimizer(self, **kwargs)

        # TRAINING LOOP
        verbose = kwargs["verbose"]
        maxiter = kwargs["maxiter"]
        if verbose:
            from tqdm import trange
            bar = trange(maxiter)
        else:
            bar = range(maxiter)
        for epoch in bar:
            l = self.loss()
            opt.zero_grad()
            l.backward()
            opt.step()
            self.after_step()

            if epoch % 50 == 0 and verbose:
                bar.set_description(f"loss: {l}")

    ####################################################################################################################

    @abstractmethod
    def loss(self, representation=None) -> T:
        pass

    def after_step(self) -> None:
        r"""
            Perform after-step operations, for example a projection of the parameters onto some manifold.
        """
        pass



"""
Custom optimizer for RKMs

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021
"""
from __future__ import annotations

import torch
from .stiefel import stiefel_optimizer

from .. import utils
from kerch.feature.feature import Module


class Optimizer():
    @staticmethod
    def param_state(params):
        out = torch.nn.ParameterList([])
        for p in params:
            if p.requires_grad is True:
                out.append(p)
        return out

    @utils.kwargs_decorator({"stiefel_lr": 1.e-3,
                             "euclidean_lr": 1.e-3,
                             "slow_lr": 1.e-4
                             })
    def __init__(self, module: Module, type="sgd", **kwargs):
        self._kwargs = kwargs
        self._type = type
        self._module = module

        euclidean_params = Optimizer.param_state(module.manifold_parameters(type='euclidean'))
        stiefel_params = Optimizer.param_state(module.manifold_parameters(type='stiefel'))
        slow_params = Optimizer.param_state(module.manifold_parameters(type='slow'))

        self._dict = []
        self._opt = None
        self._requires_grad = False

        if len(euclidean_params) > 0:
            dict_euclidean = {'params': euclidean_params, 'stiefel': False, 'lr': kwargs['euclidean_lr']}
            self._dict.append({**kwargs, **dict_euclidean})
            self._requires_grad = True

        if len(slow_params) > 0:
            dict_slow = {'params': slow_params, 'stiefel': False, 'lr': kwargs['slow_lr']}
            dict_slow = {**kwargs, **dict_slow}
            self._dict.append(dict_slow)
            self._requires_grad = True

        if len(stiefel_params) > 0:
            dict_stiefel = {'params': stiefel_params, 'stiefel': True, 'lr': kwargs['stiefel_lr']}
            self._dict.append({**kwargs, **dict_stiefel})
            self._requires_grad = True

        if self._dict:
            opt_switcher = {"sgd": stiefel_optimizer.SGDG,
                            "adam": stiefel_optimizer.AdamG}
            self._opt = opt_switcher.get(type, "Incorrect optimizer name.")(self._dict)

    @property
    def requires_grad(self):
        return self._requires_grad

    @property
    def type(self):
        return self._type

    @property
    def hparams(self):
        return {'[Optimizer] ' + key: val for key, val in self._kwargs.items()}

    @property
    def module(self) -> Module:
        return self._module

    def reduce(self, rate):
        if self._opt is not None:
            for params in self._opt.param_groups:
                params['lr'] *= rate

    def step(self, closure=None) -> float | None:
        if self._opt is not None: loss = self._opt.step(closure)
        else: loss = closure()
        return loss

    def zero_grad(self):
        if self._opt is not None: self._opt.zero_grad()



# coding=utf-8
from ._parser import get_args, get_kwargs
from .. import data, model, train


def run(**kwargs):
    verbose = kwargs.pop('verbose')
    filename = kwargs.pop('filename')
    expe_kwargs = get_kwargs(filename)
    expe_name = filename[:-len('.yaml')]

    # LOADING DATA
    if verbose: print('Loading data...', end=" ")
    dataset = data.factory(**expe_kwargs['data'])

    # LOADING MODEL
    if verbose: print('Done'), print('Loading model...', end=" ")
    mdl = model.factory(**expe_kwargs['model'])

    # LOADING TRAINER
    if verbose: print('Done'), print('Initializing training...', end=" ")
    trainer = train.Trainer(model=mdl, learning_set=dataset, verbose=verbose, expe_name=expe_name,
                            **expe_kwargs['train'])

    # TRAIN
    if verbose: print('Done'), print('Start training...')
    trainer.fit()
    if verbose: print('Model trained')


if __name__ == '__main__':
    kwargs = get_args()
    run(**kwargs)



# coding=utf-8

# ABSTRACT
from ._base_kernel import _BaseKernel
from .kernel import Kernel as Kernel
from .implicit import Implicit as Implicit
from .explicit import Explicit as Explicit
from .distance.distance import Distance as Distance
from .distance.distance_squared import DistanceSquared as DistanceSquared
from .distance.select_distance import SelectDistance as SelectDistance

# GENERIC
from .generic.linear import Linear as Linear
from .generic.rbf import RBF as RBF

from .generic.laplacian import Laplacian as Laplacian
from .generic.cosine import Cosine as Cosine
from .generic.sigmoid import Sigmoid as Sigmoid
from .generic.polynomial import Polynomial as Polynomial


# RANDOM FEATURES
from .random_features.random_features import RandomFeatures as RandomFeatures
from .random_features.rff import RFF as RFF
from .random_features.rf_lrelu import RFLReLU as RFLReLU
from .random_features.rf_arcsinh import RFArcsinh as RFArcsinh
from .random_features.rf_hyperbola import RFHyperbola as RFHyperbola
from .random_features.rf_stacked import RFStacked as RFStacked

# NETWORK
from .network.explicit_nn import ExplicitNN as ExplicitNN
from .network.implicit_nn import ImplicitNN as ImplicitNN

# TIME
from .time.indicator import Indicator as Indicator
from .time.hat import Hat as Hat

# VISION
from .vision.additive_chi_2 import AdditiveChi2 as AdditiveChi2
from .vision.skewed_chi_2 import SkewedChi2 as SkewedChi2

# STATISTICS
from .statistics.epanechnikov import Epanechnikov as Epanechnikov
from .statistics.uniform import Uniform as Uniform
from .statistics.triangular import Triangular as Triangular
from .statistics.quartic import Quartic as Quartic
from .statistics.triweight import Triweight as Triweight
from .statistics.tricube import Tricube as Tricube
from .statistics.logistic import Logistic as Logistic
from .statistics.silverman import Silverman as Silverman
from .statistics.exponential import Exponential as Exponential

# MISC
from .nystrom import Nystrom as Nystrom
from ._factory import factory as factory


"""
File containing the abstract kernel classes.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021
"""

import torch

from abc import ABCMeta, abstractmethod
from torch import Tensor

from . import utils
from ._cache import _Cache
from ._logger import _Logger

class _StochasticElement():
    pass


@utils.extend_docstring(_Cache)
class _Stochastic(_Cache,  # creates a transportable cache (e.g. for GPU)
                  metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        super(_Stochastic, self).__init__(*args, **kwargs)
        self._num_total = None
        self._idx_stochastic = None

    @property
    def num_idx(self) -> int:
        r"""
        Number of selected indices when performing various operations. This is only relevant in the case of stochastic
        training.
        """
        return len(self._idx_stochastic)

    @property
    def idx(self) -> Tensor:
        r"""
        Indices used when performing various operations. This is only relevant in the case of stochastic training.
        """
        return self._idx_stochastic

    @property
    def _all_idx(self):
        return range(self._num_total)

    def train(self, mode=True):
        r"""
        Activates the training mode, which disables the gradients computation and disables stochasticity. For the
        gradients and other things, we refer to the `torch.nn.Module` documentation. For the stochastic part, when put
        in evaluation mode (`False`), all the sample points are used for the computations, regardless of
        the previously specified indices.
        """
        if not mode:
            self.stochastic()
        return self

    def stochastic(self, idx=None, prop=None):
        """
        Resets which subset of the samples are to be used until the next call of this function. This is relevant in the
        case of stochastic training.

        :param idx: Indices of the sample subset relative to the original sample set., defaults to `None`
        :type idx: int[], optional
        :param prop: Instead of giving indices, passing a proportion of the original sample set is also
            possible. The indices will be uniformly randomly chosen without replacement. The value must be chosen
            such that :math:`0 <` `prop_stochastic` :math:`\leq 1`., defaults to `None`.
        :type prop: double, optional

        If `None` is specified for both `idx_stochastic` and `prop_stochastic`, all samples are used and the subset equals the
        original sample set. This is also the default behavior if this function is never called, nor the parameters
        specified during initialization.

        .. note::
            Both `idx_stochastic` and `prop_stochastic` cannot be filled together as conflict would arise.
        """
        if self._num_total is None:
            self._log.info("Setting stochastic indices cannot work before any value or dimensions have been fed.")
            return

        self._reset_cache()
        assert idx is None or prop is None, "Both idx_stochastic and prop_stochastic are not None. " \
                                                          "Please choose one non-None parameter only."

        if idx is not None:
            self._log.debug("Using the provided indices for stochasticity.")
            self._idx_stochastic = idx
        elif prop is not None:
            assert prop <= 1., 'Parameter prop_stochastic: the chosen proportion cannot be greater than 1.'
            assert prop > 0., 'Parameter prop_stochastic: the chosen proportion must be strictly greater than 0.'

            self._log.debug("Randomly defining the indices.")
            n = self._num_total
            k = round(n * prop)
            perm = torch.randperm(n)
            self._idx_stochastic = perm[:k]
        else:
            self._log.debug("Using all indices.")
            self._idx_stochastic = self._all_idx

        for stochastic_module in self.children():
            if isinstance(stochastic_module, _Stochastic):
                stochastic_module.stochastic(idx=self._idx_stochastic)



"""
Abstract RKM View class.

@author: HENRI DE PLAEN
@copyright: KU LEUVEN
@license: MIT
@date: March 2021, rewritten in June 2022
"""

import torch
from torch import Tensor
from math import sqrt
from abc import ABCMeta, abstractmethod
from typing import Union

from kerch import utils
from kerch.kernel import factory, _Base
from kerch._stochastic import _Stochastic

@utils.extend_docstring(_Stochastic)
class _View(_Stochastic, metaclass=ABCMeta):
    r"""
    :param kernel: Initiates a View based on an existing kernel object. If the value is not `None`, all other
        parameters are neglected and inherited from the provided kernel., default to `None`
    :param bias: Bias
    :param bias_trainable: defaults to `False`

    :type bias: bool, optional
    :type bias_trainable: bool, optional
    """

    @utils.kwargs_decorator({
        "dim_output": None,
        "hidden": None,
        "weight": None,
        "param_trainable": True,
        "representation": "dual"
    })
    def __init__(self, *args, **kwargs):
        """
        A View is made of a kernel and primal or dual variables. This second part is handled by the daughter classes.
        """
        super(_View, self).__init__(*args, **kwargs)
        self._dim_output = kwargs["dim_output"]
        self._representation = utils.check_representation(kwargs["representation"], cls=self)

        weight = kwargs["weight"]
        hidden = kwargs["hidden"]

        # INITIATE
        self._param_trainable = kwargs["param_trainable"]
        self._hidden = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE), self._param_trainable)
        self._weight = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE), self._param_trainable)
        if weight is not None and hidden is not None:
            self._log.info("Both the hidden and the weight are set. Priority is given to the weight values.")
            self.weight = weight
        elif weight is None:
            self.hidden = hidden
        elif hidden is None:
            self.weight = weight

        self._attached_weight = None

    def _reset_hidden(self) -> None:
        self._hidden = torch.nn.Parameter(torch.empty(0, dtype=utils.FTYPE,
                                                      device=self._hidden.device),
                                          requires_grad=self._hidden.requires_grad)

    @abstractmethod
    def _reset_weight(self) -> None:
        pass

    def _init_hidden(self) -> None:
        assert self._num_total is not None, "No dataset has been initialized yet."
        assert self._dim_output is not None, "No output dimension has been provided."
        self.hidden = torch.nn.init.orthogonal_(torch.empty((self._num_total, self.dim_output),
                                                            dtype=utils.FTYPE, device=self._hidden.device))

    def _init_weight(self) -> None:
        assert self._dim_output is not None, "No output dimension has been provided."
        self.weight = torch.nn.init.orthogonal_(torch.empty((self.dim_feature, self.dim_output),
                                                            dtype=utils.FTYPE, device=self._weight.device))

    def init_parameters(self, representation=None, overwrite=True) -> None:
        """
        Initializes the model parameters: the weight in primal and the hidden values in dual.
        This is suitable for gradient-based training.

        :param representation: 'primal' or 'dual'
        :param overwrite: Does not initialize already initialized parameters if False., defaults to True
        :type representation: str, optional
        :type overwrite: bool, optional
        """
        representation = utils.check_representation(representation, self._representation, cls=self)

        def init_weight():
            if overwrite or not self._weight_exists:
                self._init_weight()

        def init_hidden():
            if overwrite or not self._hidden_exists:
                self._init_hidden()

        switcher = {"primal": init_weight,
                    "dual": init_hidden}
        switcher.get(representation)()

    ################################################################"
    @property
    @abstractmethod
    def dim_feature(self) -> int:
        """
        Dimension of the explicit feature map if relevant.
        """
        pass

    @property
    def dim_output(self) -> int:
        r"""
        Output dimension.
        """
        return self._dim_output

    @dim_output.setter
    def dim_output(self, val: int):
        self._dim_output = val
        self._reset_weight()
        self._reset_hidden()

    ##################################################################################################################
    ## HIDDEN

    @property
    def hidden(self) -> Tensor:
        """
        Hidden values.
        """
        if self._hidden_exists:
            return self._hidden.T[self.idx, :]
        self._log.debug("No hidden values have been initialized yet.")

    def update_hidden(self, val: Tensor, idx_sample=None) -> None:
        # first verify the existence of the hidden values before updating them.
        if not self._hidden_exists:
            self._log.warning("Could not update hidden values as these do not exist. "
                              "Please set the values for hidden first.")
            return

        if idx_sample is None:
            idx_sample = self._all_sample()
        self._hidden.data.T[idx_sample, :] = val.data
        self._reset_weight()

    @hidden.setter
    def hidden(self, val):
        # sets the parameter to an existing one
        if val is not None:
            if isinstance(val, torch.nn.Parameter):
                self._hidden = val
                self._num_h, self._dim_output = self._hidden.shape
            else:  # sets the value to a new one
                # to work on the stiefel manifold, the parameters are required to have the number of components as
                # as first dimension
                val = utils.castf(val, tensor=True, dev=self._hidden.device)
                self._hidden.data = val.T

                # zeroing the gradients if relevant
                if self._param_trainable and self._hidden.grad is not None:
                    self._hidden.grad.sample.zero_()

                self._dim_output, self._num_h = self._hidden.shape
                self._reset_weight()
        else:
            self._log.info("The hidden value is unset.")

    @property
    def hidden_trainable(self) -> bool:
        """
        Returns whether the hidden variables are trainable (a gradient can be computed on it).
        """
        return self._param_trainable

    @hidden_trainable.setter
    def hidden_trainable(self, val: bool):
        # changes the possibility of training the hidden values through backpropagation
        self._param_trainable = val
        self._hidden.requires_grad = self._param_trainable

    @property
    def _hidden_exists(self) -> bool:
        r"""
        Returns if this View has hidden variables attached to it.
        """
        return self._hidden.nelement() != 0

    ####################################################################################################################
    ## ATTACH
    @property
    def attached(self) -> bool:
        r"""
        Boolean indicating whether the view is attached to another multiview.
        """
        return self._attached_weight is not None

    def attach_to(self, weight_fn) -> None:
        self._log.debug(self.__repr__() + " is attached to a multi-view.")
        assert not self.attached, 'Cannot attach a view which is already attached'
        self._attached_weight = weight_fn

    def detach(self) -> None:
        self._log.debug(self.__repr__() + ' is now detached.')
        w = self.weight
        self._attached = False
        self.weight = w

    ## WEIGHT
    @property
    def weight(self) -> Tensor:
        r"""
        Weight.
        """
        if self.attached:
            return self._attached_weight()
        else:
            if self._weight_exists:
                return self._weight.T
            self._log.debug("No weight has been initialized yet.")

    @weight.setter
    def weight(self, val):
        if self.attached:
            raise Exception("Cannot assign value of an attached view. Please change the value of the mother class "
                            "instead.")
        else:
            if val is not None:
                # sets the parameter to an existing one
                if isinstance(val, torch.nn.Parameter):
                    self._weight = val
                else:  # sets the value to a new one
                    val = utils.castf(val, tensor=False, dev=self._weight.device)
                    if self._weight_exists:
                        self._weight.data = val.T
                        # zeroing the gradients if relevant
                        if self._param_trainable and self._weight.grad is not None:
                            self._weight.grad.sample.zero_()
                    else:
                        self._weight = torch.nn.Parameter(val.T, requires_grad=self._param_trainable)

                    self._dim_output = self._weight.shape[0]
                self._reset_hidden()
            else:
                self._log.info("The weight is unset.")

    @property
    def _weight_exists(self) -> bool:
        try:
            return self._weight.nelement() != 0
        except AttributeError:
            return False

    @abstractmethod
    def _update_weight_from_hidden(self):
        pass

    @abstractmethod
    def _update_hidden_from_weight(self):
        pass

    ## MATHS

    @abstractmethod
    def phi(self, x=None) -> Tensor:
        pass

    @abstractmethod
    def k(self, x=None) -> Tensor:
        pass

    def c(self, x=None) -> Tensor:
        phi = self.phi(x)
        return phi.T @ phi / self.num_idx

    def h(self, x=None) -> Union[Tensor, torch.nn.Parameter]:
        if not self._hidden_exists and self._weight_exists:
            try:
                self._update_hidden_from_weight()
            except NotImplementedError:
                self._log.info('The relation between the weights and the hidden variables have '
                               'not been implemented in this case.')

        if x is None:
            if self._hidden_exists:
                return self.hidden
            else:
                self._log.warning("No hidden values exist or have been initialized.")
                raise utils.ImplicitError(self)
        raise NotImplementedError

    def w(self, x=None) -> Union[Tensor, torch.nn.Parameter]:
        if not self.attached and not self._weight_exists and self._hidden_exists:
            self._update_weight_from_hidden()

        if x is None:
            return self.weight
        raise NotImplementedError

    def phiw(self, x=None, representation="dual") -> Tensor:
        def primal(x):
            return self.phi(x) @ self.W

        def dual(x):
            return self.kernel.k(x) @ self.H

        switcher = {"primal": primal,
                    "dual": dual}
        if representation in switcher:
            return switcher.get(representation)(x)
        else:
            raise utils.RepresentationError

    @property
    def Phi(self) -> Tensor:
        return self.phi()

    @property
    @abstractmethod
    def K(self) -> Tensor:
        pass

    @property
    def C(self) -> Tensor:
        return self.c()

    @property
    def H(self) -> Union[Tensor, torch.nn.Parameter]:
        return self.h()

    @property
    def W(self) -> Union[Tensor, torch.nn.Parameter]:
        return self.w()

    @property
    def PhiW(self) -> Tensor:
        return self.phiw()

    @abstractmethod
    def forward(self, x=None, representation="dual"):
        pass



from torch import Tensor as T
from abc import ABCMeta, abstractmethod

from ._level import _Level
from .view import View
from kerch import utils


class Level(_Level, View, metaclass=ABCMeta):

    @utils.extend_docstring(_Level)
    @utils.extend_docstring(View)
    def __init__(self, *args, **kwargs):
        super(Level, self).__init__(*args, **kwargs)

    ####################################################################################################################

    def solve(self, sample=None, target=None, representation=None, **kwargs) -> None:
        r"""
        Fits the model according to the input ``sample`` and output ``target``. Many models have both a primal and
        a dual formulation to be fitted.

        :param sample: Input sample of the model., defaults to the sample provided by the model.
        :param target: Target sample of the model, defaults to ```None``
        :param representation: Representation of the model (``"primal"`` or ``"dual"``)., defaults to ``"dual"``.

        :type sample: Matrix, optional
        :type target: Matrix or vector, optional
        :type representation: str, optional
        """

        # set the sample to input (always works for the underlying kernel)
        if sample is not None:
            self._log.info("Setting the sample to the provided input. Possibly overwriting a previous one.")
            self.init_sample(sample)  # keeping the stochastic state if set.

        # verify that the output has the same dimensions
        if target is not None:
            target = utils.castf(target, tensor=True)
            same_dim = sample.shape[0] == target.shape[0]
            if not same_dim:
                self._log.error("The number of sample points is not consistent with the output dimensions")
                return

        # solve model
        return super(Level, self).solve(sample=sample,
                                        target=target,
                                        representation=representation,
                                        **kwargs)

    ####################################################################################################################

    @abstractmethod
    def loss(self, representation=None) -> T:
        pass



from .Model import Model

class KPCA(Model):
    def __init__(self, *args, **kwargs):
        super(KPCA, self).__init__(*args, **kwargs)
        kwargs['level_type'] = 'kpca'
        self._add_level(*args, **kwargs)


# coding=utf-8
from ._LearningSet import _LearningSet
def factory(dataset='gaussians', **kwargs) -> _LearningSet:
    r"""
    # TODO
    """
    def case_insensitive_getattr(obj, attr):
        for a in dir(obj):
            if a.lower() == attr.lower():
                return getattr(obj, a)
        return None

    dataset = dataset.replace("_","")
    dataset = dataset.replace(" ","")

    if dataset.lower() == "pima":
        dataset = "pimaindians"

    import kerch.data
    learning_set = case_insensitive_getattr(kerch.data, dataset)
    if learning_set is None:
        raise NameError("Invalid dataset.")

    return learning_set(**kwargs)



# coding=utf-8
import os
import numpy as np
import torch

from ..utils import kwargs_decorator, FTYPE
from ._LearningSet import _LearningSetTrain, _LearningSetTrainTest

def _get_file_path(name: str):
    dir_path = os.path.dirname(__file__)
    return os.path.join(dir_path, 'files', name)


class TwoMoons(_LearningSetTrain):
    def __init__(self, *args, **kwargs):
        self._noise = kwargs.pop('noise', .1)
        self._separation = kwargs.pop('separation', [1, .5])
        assert isinstance(self._noise, (int, float))
        assert self._noise >= 0.
        assert isinstance(self._separation, (list, tuple))
        assert len(self._separation) == 2

        # range computation
        min_x = min(-1, self._separation[0]-1) - 2. * self._noise
        max_x = max(1, self._separation[0]+1) + 2. * self._noise
        min_y = min(0, self._separation[1]-1) - 2. * self._noise
        max_y = max(1, self._separation[1]) + 2. * self._noise

        fact_x = (max_x - min_x) / 2
        fact_y = (max_y - min_y) / 2

        range = (min_x - fact_x, max_x + fact_x,
                 min_y - fact_y, max_y + fact_y)

        super(TwoMoons, self).__init__(name="Two Moons",
                                       dim_data=2,
                                       dim_labels=1,
                                       range=range,
                                       **kwargs)


    def _training(self, num):
        n_samples_out = num // 2
        n_samples_in = num - n_samples_out

        outer_circ_x = torch.cos(torch.linspace(0, torch.pi, n_samples_out))
        outer_circ_y = torch.sin(torch.linspace(0, torch.pi, n_samples_out))
        inner_circ_x = self._separation[0] - torch.cos(torch.linspace(0, torch.pi, n_samples_in))
        inner_circ_y = self._separation[1] - torch.sin(torch.linspace(0, torch.pi, n_samples_in))

        data = torch.cat((torch.stack([inner_circ_x, inner_circ_y], dim=1),
                          torch.stack([outer_circ_x, outer_circ_y], dim=1)))
        labels = torch.cat((torch.ones(n_samples_in, dtype=FTYPE), torch.zeros(n_samples_out, dtype=FTYPE)))[:, None]

        if self._noise != 0.:
            data += self._noise * torch.randn(data.shape)

        return data, labels


class PimaIndians(_LearningSetTrain):
    def __init__(self, *args, **kwargs):
        super(PimaIndians, self).__init__(name="Pima Indian Diabetes",
                                          dim_data=8,
                                          dim_labels=1,
                                          **kwargs)

    def _training(self, num):
        path = _get_file_path('pima.csv')
        diabetes = np.genfromtxt(path, delimiter=',')
        data = diabetes[:, :self._dim_data]
        labels = diabetes[:, self._dim_data:]
        labels = np.where(labels == 0, -1, 1)
        return data, labels


class Spiral(_LearningSetTrain):
    def __init__(self, *args, **kwargs):
        super(Spiral, self).__init__(name="Spiral",
                                      dim_data=2,
                                      dim_labels=1,
                                     range=(-5, 6, -5, 5),
                                     **kwargs)

    def _training(self, num):
        import math
        size = -(num // -2) # ceil integer division

        def spiral_xy(i, spiral_num):
            """
            Create the value for a spiral.

            Arguments:
                i runs from 0 to 96
                spiral_num is 1 or -1
            """
            phi = i / 16 * math.pi
            r = 70 * ((104 - i) / 104)
            x = (r * math.cos(phi) * spiral_num) / 13 + 0.5
            y = (r * math.sin(phi) * spiral_num) / 13 + 0.5
            return (x, y)

        def spiral(spiral_num):
            return [spiral_xy(i, spiral_num) for i in range(size)]

        s1 = spiral(1)
        s2 = spiral(-1)

        data = np.concatenate((s1, s2))
        labels = np.concatenate((np.repeat(-1, size), np.repeat(1, size)))
        return data, labels

class Gaussians(_LearningSetTrain):
    def __init__(self, *args, **kwargs):
        super(Gaussians, self).__init__(name="Gaussians",
                                        dim_data=2,
                                        dim_labels=1,
                                        **kwargs)

    def _training(self, num):
        size = -(num // -2) # ceil integer division
        s1, m1 = .7, (2, 1)
        s2, m2 = 1.2, (-2, -3)

        g1 = np.random.normal(m1, s1, (size, 2))
        g2 = np.random.normal(m2, s2, (size, 2))

        data = np.concatenate((g1, g2))
        labels = np.concatenate((np.repeat(-1, size), np.repeat(1, size)))
        return data, labels

class Iris(_LearningSetTrain):
    def __init__(self, *args, **kwargs):
        super(Iris, self).__init__(name="Fisher's Iris",
                                   dim_data=4,
                                   dim_labels=3,
                                   **kwargs)

    def _training(self, num):
        path = _get_file_path('iris.csv')
        diabetes = np.genfromtxt(path, delimiter=',')
        data = diabetes[:, :self._dim_data]
        labels = diabetes[:, self._dim_data:]
        return data, labels




# coding=utf-8
from typing import Union, List

import kerch
import torch
from torch import Tensor
from torch.nn import Parameter

from .Transform import Transform
from .all import (UnitSphereNormalization,
                  MinimumCentering,
                  MeanCentering,
                  UnitVarianceNormalization,
                  MinMaxNormalization)


@kerch.utils.extend_docstring(Transform)
class TransformTree(Transform):
    r"""
    Creates a tree of transform for efficient computing, with cache management.

    :param explicit: True is the transform are to be computed in the explicit formulation, False instead.
    :param sample: Default sample on which the statistics are to be computed.
    :param default_transform: Optional default list of transform., defaults to None.
    :param diag_fun: Optional function handle to directly compute the diagonal of the implicit formulation to increase
        computation speed.

    :type explicit: bool
    :type sample: Tensor or function handle
    :type default_transform: List[str]
    :type diag_fun: Function handle
    """

    _all_transform = {"normalize": UnitSphereNormalization,  # for legacy
                      "center": MeanCentering,  # for legacy
                      "sphere": UnitSphereNormalization,
                      "min": MinimumCentering,
                      "variance": UnitVarianceNormalization,
                      "standard": [MeanCentering, UnitVarianceNormalization],
                      "minmax": MinMaxNormalization,
                      "unit_sphere_normalization": UnitSphereNormalization,
                      "mean_centering": MeanCentering,
                      "minimum_centering": MinimumCentering,
                      "unit_variance_normalization": UnitVarianceNormalization,
                      "minmax_normalization": MinMaxNormalization,
                      "standardize": [MeanCentering, UnitVarianceNormalization],
                      "minmax_rescaling": [MinimumCentering, MinMaxNormalization]}

    @staticmethod
    def beautify_transform(transform: list[str]) -> Union[None, List[Transform]]:
        r"""
        Creates a list of _Transform classes and removes duplicates.

        :param transform: list of the different transform.
        :type transform: List[str]
        """
        if transform is None:
            return None
        else:
            transform_classes = []
            for tr in transform:
                try:
                    if issubclass(tr, Transform):
                        transform_classes.append(tr)
                except TypeError:
                    new_transform = TransformTree._all_transform.get(
                        tr, NameError(f"Unrecognized transform key {tr}."))
                    if isinstance(new_transform, Exception):
                        raise new_transform
                    elif isinstance(new_transform, List):
                        for ntr in new_transform:
                            transform_classes.append(ntr)
                    elif issubclass(new_transform, Transform):
                        transform_classes.append(new_transform)
                    else:
                        kerch._GLOBAL_LOGGER._logger.error("Error while creating TransformTree list of transform")

            # remove same following elements
            previous_item = None
            idx = 0
            for current_item in transform_classes:
                if current_item == previous_item:
                    transform_classes.pop(idx)
                else:
                    previous_item = current_item
                    idx += 1

            return transform_classes

    def __init__(self, explicit: bool, sample, default_transform=None, diag_fun=None, **kwargs):
        super(TransformTree, self).__init__(explicit=explicit, name='base', **kwargs)

        if default_transform is None:
            default_transform = []

        self._default_transforms = TransformTree.beautify_transform(default_transform)

        # create default tree
        node = self
        for transform in self._default_transforms:
            offspring = transform(explicit=self.explicit, default_path=True, cache_level=self.cache_level)
            node.add_offspring(offspring)
            node = offspring
        self._default_node = node
        node.default = True

        self._base = sample
        self._data_oos = None

        self._diag_fun = diag_fun

    def __str__(self):
        output = "Transforms: \n"
        if len(self._default_transforms) == 0:
            return output + "\t" + "None (default)"
        node = self._default_node
        while not isinstance(node, TransformTree):
            output += "\t" + str(node)
            node = node.parent
        return output

    @property
    def default_transforms(self) -> List:
        r"""
        Default list of transforms to be applied.
        """
        return self._default_transforms

    @property
    def final_transform(self) -> type(Transform):
        r"""
        Final transform to be applied, which is the last element of
        :py:attr:`~kerch.transform.TransformTree.default_transforms`.
        """
        try:
            return self._default_transforms[-1]
        except IndexError:
            return TransformTree

    def _get_data(self) -> Union[Tensor, Parameter]:
        if callable(self._base):
            return self._base()
        return self._base

    def _explicit_statistics(self, sample):
        return None

    def _implicit_statistics(self, sample, x=None):
        return None

    def _explicit_sample(self):
        if callable(self._base):
            return self._base()
        return self._base

    def _implicit_sample(self):
        if callable(self._base):
            return self._base()
        return self._base

    def _implicit_diag(self, x=None) -> Union[Tensor]:
        if self._diag_fun is not None:
            return self._diag_fun(x)
        if x is None:
            return torch.diag(self._implicit_sample())[:, None]
        else:
            return torch.diag(self._implicit_oos(x, x))[:, None]

    @property
    def projected_sample(self) -> Tensor:
        r"""
        Sample after transform. Retrieved from cache if relevant.
        """
        return self._default_node.sample

    def _explicit_statistics_oos(self, oos, x=None):
        pass

    def _implicit_statistics_oos(self, oos, x=None):
        pass

    def _explicit_oos(self, x=None):
        if callable(self._base):
            return self._base(x)
        return x

    def _implicit_oos(self, x=None, y=None):
        if callable(self._base):
            return self._base(x, y)
        raise NotImplementedError

    def _revert_explicit(self, oos):
        return oos

    def _revert_implicit(self, oos):
        return oos

    def _get_tree(self, transform: List[str] = None) -> List[Transform]:
        transform = TransformTree.beautify_transform(transform)
        if transform is None:
            transform = self._default_transforms

        tree_path = [self]
        for tr_class in transform:
            current_tr = tree_path[-1]
            if tr_class in current_tr.offspring:
                offspring = current_tr.offspring[tr_class]
            else:
                offspring = tr_class(explicit=self.explicit, cache_level=self.cache_level)
                current_tr.add_offspring(offspring)
            tree_path.append(offspring)
        return tree_path

    def apply(self, oos=None, x=None, y=None, transform: List[str] = None) -> Tensor:
        r"""
        Applies the transform to the value to out-of-sample data. Either value is a function handle and you can use
        x (explicit) and x, y (explicit) to specify the data. Either directly give a Tensor.

        .. warning::
            If value is a Tensor, some transform may not work in implicit formulation. For example, the unit sphere
            normalization requires k(x,x) for all out-of-sample points. Some combinations may be even more intricate.

        :param x: Relevant if using a function handle for value.
        :param y: Relevant if using a function handle for value in implicit mode.
        :param transform: Transforms to be used. If none are to be used, i.e. getting the raw data back, please
            specify [], not None, which will return the default transform used for the sample., defaults to None,
            i.e., the default transform.

        :type x: Tensor
        :type y: Tensor
        :type transform: List[str]
        """
        tree_path = self._get_tree(transform)
        sol = tree_path[-1].oos(x=x, y=y)
        self._clean_cache()
        return sol

    def revert(self, value, transform: List[str] = None) -> Tensor:
        r"""
        Reverts the transform (runs the tree backwards) to the value to out-of-sample data.

        :param value: Out-of-sample data.
        :param transform: Transforms to be used. If none are to be used, i.e. getting the raw data back, please
            specify [], not None, which will return the default transform used for the sample., defaults to None,
            i.e., the default transform.

        :type value: Tensor
        :type transform: List[str]
        """

        tree_path = self._get_tree(transform)
        for transform in reversed(tree_path):
            value = transform._revert(value)
        return value



"""
Abstract class allowing for torch.nn.Modules to also host a cache of torch objects
which can also be ported like torch.nn.Parameters are, on GPU for example.
"""

import torch
from typing import Union
from abc import ABCMeta, abstractmethod

from ._module import _Module
from .utils import kwargs_decorator, extend_docstring


@extend_docstring(_Module)
class _Cache(_Module,
             metaclass=ABCMeta):
    r"""
    :param cache_level: level for a cache, defaults to 'normal'.
    :type cache_level: str
    """

    cache_level_switcher = switcher = {"oblivious": 0,
                                       "lightweight": 1,
                                       "normal": 2,
                                       "heavy": 3,
                                       "total": 4}

    @abstractmethod
    @kwargs_decorator({
        "cache_level": "normal"
    })
    def __init__(self, *args, **kwargs):
        super(_Cache, self).__init__(*args, **kwargs)

        # we initiate the cache
        self._cache = {}
        self.cache_level = kwargs["cache_level"]

    @property
    def cache_level(self) -> str:
        r"""
        Cache level possible values (defaults to "normal"):
        * "oblivious": the cache is inexistent and everything is computed on the go.
        * "lightweight": the cache is very light. For example, only the kernel matrix and statistics of the sample
            points are saved.
        * "normal": same as lightweight, but the statistics of the out-of-sample points are also saved.
        * "heavy": in addition to the statistics, the final kernel matrices of the out-of-sample points are saved.
        * "total": every step of any computation is saved.
        """
        switcher = _Cache.cache_level_switcher
        inv_switcher = {switcher[k]: k for k in switcher}
        return inv_switcher[self._cache_level]

    @cache_level.setter
    def cache_level(self, val: Union[str, int]):
        val = self._get_level(val)
        self._cache_level = val

    def _get_level(self, level: Union[int, str, None]) -> int:
        r"""
        Transforms the cache level to an int if not already.
        """
        if level is None:
            return self._cache_level
        if isinstance(level, str):
            level = _Cache.cache_level_switcher.get(level, "Unrecognized cache level.")
        return level

    def _apply(self, fn):
        r"""
        This if the native function by torch.nn.Module when ported. Here, the sole porting
        of the cache is also added. This is used for example to port the data to the gpu or cpu.
        """
        with torch.no_grad():
            for value in self._cache.values():
                cache_entry = value[1]
                if isinstance(cache_entry, torch.Tensor):
                    cache_entry.data = fn(cache_entry)
                elif isinstance(cache_entry, _Cache):
                    cache_entry._apply(fn)
        return super(_Cache, self)._apply(fn)

    def _get(self, key, level, fun, force: bool = False, overwrite: bool = False):
        r"""
        Retrieves an element from the cache. If the element is not present, it saved to the cache provided its level
        is lower or equal to the default level. This can be overwritten by the overwrite argument.

        :param name: key of the cache element
        :param level: level where to save the cache element
        :param fun: function to compute the element
        :param force: if the value is True, the element will nevertheless be saved whatever level
        :param overwrite: forces the recomputation of the memory element and potential overwriting

        :type key: str
        :type level: str
        :type fun: function handle
        :type force: bool
        :type overwrite: bool
        """
        level = self._get_level(level)

        if not overwrite:
            if key in self._cache:
                return self._cache[key][1]
            elif type(key) is tuple:
                reverted_key = (key[1],key[0])
                if reverted_key in self._cache:
                    return self._cache[reverted_key][1].T

        val = fun()
        if level <= self._cache_level or force:
            self._cache[key] = (level, val)
        return val

    def _reset_cache(self) -> None:
        r"""
        this just resets the cache and makes it empty. If refilled, the new elements
        will be on the same support as the rest of the module as created by its parameters.
        """

        self._log.debug(f"The cache is resetted.")
        # for val in self._cache.values(): del val
        self._cache = {}

    def _clean_cache(self, max_level: Union[str, int, None] = None):
        r"""
        Cleans all cache elements that have been forced.

        :param max_level: all levels above this level will be cleaned, max_level excluded. Defaults to the default
            cache level.
        """
        max_level = self._get_level(max_level)
        self._log.debug(f"The cache is cleaned for levels {max_level} and above).")
        # for val in self._cache.values(): del val
        if max_level == 0:
            self._cache = {}
        else:
            for key, value in self._cache.items():
                if value[0] > max_level:
                    del self._cache[key]

    def _remove_from_cache(self, key: str) -> None:
        r"""
        Removes a specific element from the cache.

        :param key: Key of the cache element to be removed
        """
        if key in self._cache:
            del self._cache[key]

    def reset(self, children=False) -> None:
        r"""
        Resets the cache to empty.

        :param children: TODO
        """
        self._reset_cache()
        for cache in self.children():
            if isinstance(cache, _Cache):
                cache.reset(children=children)



# coding=utf-8
import torch
from ..feature.logger import _GLOBAL_LOGGER


def eigs(A, k=None, B=None, psd=True, sym=True):
    r"""
    Eigenvalue decomposition. This method is a wrapper calling other methods depending on the context. In a kernel 
    context, most matrices are symmetric because kernels also are. Hence, they are Hermitian and a faster SVD can be 
    used. Alternatively, all the eigenvalues are not necessarily required and we may thus skip the computation of a 
    full eigendecomposition. The goal of this method is to always choose the most efficient method depending on the
    context.
    
    :param A: Matrix to be decomposed.
    :param k: Number of greatest eigenpairs requested. Defaults to `None`, which corresponds to computing all of them.
    :param B: Matrix in the case of a generalized eigenvalue problem. Specify `None` (default) for a classical
        eigenvalue decomposition.
    :param psd: Specifies whether the matrix `A` is positive semi-definite. Defaults to `True`.
    :param sym: Specifies whether the matrix `A` is positive symmetric. Defaults to `True`.
    :return: eigenvalues, eigenvectors.

    :type A: torch.Tensor
    :type k: int, optional
    :type B: torch.Tensor, optional
    :type psd: bool, optional
    :type sym: bool, optional
    :rtype: Tuple[torch.Tensor, torch.Tensor]
    """
    assert A is not None, 'Cannot decompose an empty matrix.'
    k1, k2 = A.shape
    assert k1 == k2, f'This function can only decompose square matrices (found {k1}x{k2}).'
    if k is None: k = k1
    assert k <= k1, f'Requested eigenvectors ({k}) exceeds matrix dimensions ({k1}).'

    try:
        s, v = torch.lobpcg(A, k=k, B=B, largest=True)
        _GLOBAL_LOGGER._logger.info('Using LOBPCG for eigendecomposition.')
    except:
        if sym:
            if B is None:
                s, v = torch.linalg.eigh(A)
            else:
                s, v = torch.linalg.eigh(torch.linalg.inv(B) @ A)
            _GLOBAL_LOGGER._logger.info('Using hermitian eigendecomposition (eigh).')
            v = v[:, -k:]  # eigenvectors are vertical components of v
            s = s[-k:]
            v = torch.flip(v, dims=(1,))
            s = torch.flip(s, dims=(0,))
        elif psd:
            if B is None:
                _, s, v = torch.svd(A)
            else:
                _, s, v = torch.svd(torch.linalg.inv(B) @ A)
            _GLOBAL_LOGGER._logger.info('Using SVD for eigendecomposition (svd).')
            v = v[:, :k]  # eigenvectors are vertical components of v
            s = s[:k]
        else:
            if B is None:
                s, v = torch.linalg.eig(A)
            else:
                s, v = torch.linalg.eig(torch.linalg.inv(B) @ A)
            _GLOBAL_LOGGER._logger.info('Using classical eigendecomposition (eig).')
            v = v[:, :k]  # eigenvectors are vertical components of v
            s = s[:k]

    return s.data, v.data



# coding=utf-8
from __future__ import annotations

import torch
from torch.utils.data import SequentialSampler
import tqdm

from ..model.Model import Model
from ..feature.logger import Logger
from ..utils import kwargs_decorator
from ..opt import Optimizer
from ..data import _LearningSet
from .. import monitor


class Trainer(Logger):
    def __init__(self, model: Model, learning_set: _LearningSet, **kwargs):
        super(Trainer, self).__init__(**kwargs)
        self._learning_set: _LearningSet = learning_set
        self._model: Model = model
        self.problem = kwargs.pop('problem', 'regression')
        self.batch_size = kwargs.pop('batch_size', 'all')
        self.epochs = kwargs.pop('epochs', 100)
        self.use_gpu = kwargs.pop('use_gpu', False)
        self.stiefel_lr = kwargs.pop('stiefel_lr', 1e-3)
        self.euclidean_lr = kwargs.pop('euclidean_lr', 1e-3)
        self.slow_lr = kwargs.pop('slow_lr', 1e-4)
        self._verbose = kwargs.pop('verbose', False)
        self._watcher_kwargs = kwargs.pop('watcher', dict())
        self._expe_name = kwargs.pop('expe_name', 'expe')

    @property
    def problem(self) -> str:
        return self._problem

    @problem.setter
    def problem(self, val: str):
        msg = "The problem argument can either be set to the strings 'regression' or 'classification'."
        assert isinstance(val, str), msg
        if val == 'regression':
            self._loss = torch.nn.MSELoss(reduction='mean')
        elif val == 'classification':
            raise NotImplementedError
        else:
            raise AssertionError(msg)
        self._problem = val

    @property
    def model(self) -> Model:
        return self._model

    @property
    def epochs(self) -> int:
        return self._epochs

    @epochs.setter
    def epochs(self, val: int):
        assert val > 0, "The number of epochs must be strictly positive."
        self._epochs = val

    @property
    def batch_size(self) -> int | str:
        return self._batch_size

    @batch_size.setter
    def batch_size(self, val: int | str):
        msg = "The batch_size argument can only take an integer or the string 'all' as valid option."
        num = len(self._learning_set.training_set)
        if isinstance(val, str):
            assert val == 'all', msg
            self._sampler = torch.utils.data.BatchSampler(SequentialSampler(range(num)),
                                                          batch_size=num, drop_last=False)
        else:
            assert isinstance(val, int), msg
            assert val > 0, \
                f"Only positive values for the batch size are allowed ({val} given)."
            assert val <= num, \
                f"The batch size ({val}) cannot be greater than the number of elements in the training set ({num})."
            self._sampler = torch.utils.data.BatchSampler(SequentialSampler(range(num)),
                                                          batch_size=val, drop_last=False)
        self._batch_size = val

    @property
    def _mask_batch_progress(self) -> bool:
        return self.batch_size == 'all'

    @property
    def _training_data(self) -> torch.Tensor:
        return (self._learning_set.training_set[:])[0]

    @property
    def _training_labels(self) -> torch.Tensor:
        return (self._learning_set.training_set[:])[1]

    @property
    def _validation_data(self) -> torch.Tensor:
        return (self._learning_set.validation_set[:])[0]

    @property
    def _validation_labels(self) -> torch.Tensor:
        return (self._learning_set.validation_set[:])[1]

    @property
    def _test_data(self) -> torch.Tensor:
        return (self._learning_set.test_set[:])[0]

    @property
    def _test_labels(self) -> torch.Tensor:
        return (self._learning_set.test_set[:])[1]

    @property
    def use_gpu(self) -> bool:
        r"""
        Specifies whether the gpu has to be used
        """
        return self._use_gpu

    @use_gpu.setter
    def use_gpu(self, val: bool):
        self._use_gpu = val

    @property
    def stiefel_lr(self) -> float:
        r"""
        Learning rate for the Stiefel manifold
        """
        return self._stiefel_lr

    @stiefel_lr.setter
    def stiefel_lr(self, val: float):
        self._stiefel_lr = val

    @property
    def euclidean_lr(self) -> float:
        r"""
        Learning rate for the Euclidean manifold, i.e. the 'normal' parameters.
        """
        return self._euclidean_lr

    @euclidean_lr.setter
    def euclidean_lr(self, val: float):
        self._euclidean_lr = val

    @property
    def slow_lr(self) -> float:
        r"""
        Learning rate for specific hyperparameters that are better trained slower, i.e. the bandwidth of
        an RBF kernel.
        """
        return self._slow_lr

    @slow_lr.setter
    def slow_lr(self, val: float):
        self._slow_lr = val

    def _init_fit(self):
        if self.use_gpu:
            self.model.to('cuda:0')

        self.model.init_sample(self._training_data)
        self.model.init_levels()
        self.model.init_target(self._training_labels)

        self._optimizer = Optimizer(module=self.model,
                                    stiefel_lr=self._stiefel_lr,
                                    euclidean_lr=self._euclidean_lr,
                                    slow_lr=self._slow_lr)

    @torch.no_grad()
    def _problem_loss(self, data, labels) -> float:
        if len(data) == 0:
            return 0.
        self.model.eval()
        pred = self.model(data)
        self.model.train()
        return self._loss(pred, labels)

    def fit(self) -> Model:
        self._init_fit()
        self.model.train()

        epoch_progress = tqdm.tqdm(range(self.epochs),
                                   position=0,
                                   leave=True,
                                   desc="Epoch",
                                   disable=not self._verbose)
        batch_progress = tqdm.tqdm(self._sampler,
                                   position=0,
                                   leave=False,
                                   disable=self._mask_batch_progress or not self._verbose,
                                   desc="Batch")

        watcher = monitor.factory(model=self.model, opt=self._optimizer, expe_name=self._expe_name,
                                  verbose=self._verbose, **self._watcher_kwargs)

        def closure():
            self.model()
            loss = self._model.loss()
            if self._optimizer.requires_grad:
                loss.backward()
            return loss

        # TRAINING LOOP
        for epoch in epoch_progress:
            # preliminaries
            self._optimizer.zero_grad()
            if self.use_gpu:
                torch.cuda.empty_cache()
            batch_loss = 0.
            # forward and backward
            for idx_batch in batch_progress:
                self.model.before_step()
                self.model.stochastic(idx=idx_batch)
                batch_loss += self._optimizer.step(closure)
                self.model.after_step()

            # validation and test
            train_loss = self._problem_loss(self._training_data, self._training_labels)
            val_loss = self._problem_loss(self._validation_data, self._validation_labels)
            test_loss = self._problem_loss(self._test_data, self._test_labels)
            watcher.update(epoch=epoch,
                           objective_loss=batch_loss,
                           training_error=train_loss,
                           validation_error=val_loss,
                           test_error=test_loss)

        watcher.finish()
        self.model.eval()
        return self.model



# coding=utf-8
from __future__ import annotations
import torch
from .type import ITYPE, FTYPE
from .errors import RepresentationError

def castf(x, dev=None, tensor=True) -> torch.Tensor | None:
    r"""
    Casts the input to a PyTorch float tensor. If the input is a scalar, it is cast to a tensor. The cast can be done
    to a 1D or a 2D tensor depending on the parameter `tensor`. If the provided data x has more than 2 dimensions, an
    error is raised. The default floating type used is :attr:`kerch.FTYPE`. The `None` values are not casted and 
    returned as is.

    :param x: The input to cast.
    :param dev: The device to cast the tensor to. Defaults to `None`, which corresponds to no device change.
    :param tensor: If True, the input is cast to a 2D tensor. If False, the input is cast to a 1D tensor.
    :return: The input cast to a PyTorch float tensor, with optional device.

    :type x: float | torch.Tensor | np.ndarray | None
    :type dev: Optional[torch.device]
    :type tensor: bool
    :rtype: torch.Tensor | None
    """
    if x is None:
        return None

    if not torch.is_tensor(x):
        x = torch.tensor(x, requires_grad=False, dtype=FTYPE, device=dev)
    else:
        x = x.type(FTYPE)

    if dev is not None:
        x = x.to(dev)

    if tensor:
        dim = len(x.shape)
        if dim == 0:
            x = x.unsqueeze(0)
            dim = 1
        if dim == 1:
            x = x.unsqueeze(1)
        elif dim > 2:
            raise NameError(f"Provided data has too much dimensions ({dim}).")

    return x

def casti(x, dev=None, tensor=False) -> torch.Tensor | None:
    r"""
    Casts the input to a PyTorch integer tensor. If the input is a scalar, it is cast to a tensor. The cast can be done
    to a 1D or a 2D tensor depending on the parameter `tensor`. If the provided data x has more than 2 dimensions, an
    error is raised. The default floating type used is :attr:`kerch.ITYPE`. The `None` values are not casted and 
    returned as is.

    :param x: The input to cast.
    :param dev: The device to cast the tensor to. Defaults to `None`, which corresponds to no device change.
    :param tensor: If True, the input is cast to a 2D tensor. If False, the input is cast to a 1D tensor.
    :return: The input cast to a PyTorch integer tensor, with optional device.

    :type x: int | torch.Tensor | np.ndarray | None
    :type dev: Optional[torch.device]
    :type tensor: bool
    :rtype: torch.Tensor | None
    """
    if x is None:
        return None

    if not torch.is_tensor(x):
        x = torch.tensor(x, requires_grad=False, dtype=ITYPE, device=dev)
    else:
        x = x.type(ITYPE)

    if dev is not None:
        x = x.to(dev)

    if tensor:
        dim = len(x.shape)
        if dim == 0:
            x = x.unsqueeze(0)
            dim = 1
        if dim == 1:
            x = x.unsqueeze(1)
        elif dim > 2:
            raise NameError(f"Provided data has too much dimensions ({dim}).")

    return x.squeeze()


def check_representation(representation: str = None, default: str = None, cls=None) -> str:
    r"""
    This model verifies whether the provided representation is valid. If the representation is `None` and a default
    value is provided, the default value is returned. If the representation is not `None` and is not valid, an error is
    raised. The valid representations are `primal` and `dual`.
    
    :param representation: The representation to check.
    :param default: Default representation for the case where `representation` is `None`.
    :param cls: An instance of :class:`kerch.feature.Logger` to throw the error from, typically the one calling this
        method. This is optional.
    :return: "primal" | "dual"

    :type representation: str, optional
    :type default: str, optional
    :type cls: kerch.feature.Logger, optional
    :rtype: str
    """
    if representation is None and default is not None:
        representation = default

    valid = ["primal", "dual"]
    if representation not in valid:
        raise RepresentationError(cls)
    return representation


def capitalize_only_first(val: str) -> str:
    r"""
    This method returns the input string with the first letter capitalized and the rest of the string unchanged.

    :param val: String to be capitalized.
    :return: Capitalized string.
    :type val: str
    :rtype: str
    """
    return val[0].upper() + val[1:]



"""
Abstract class defining a general module in the toolbox.
"""

import torch
from typing import Iterator
from abc import ABCMeta, abstractmethod

from ._logger import _Logger
from .utils import capitalize_only_first, extend_docstring


@extend_docstring(_Logger)
class _Module(_Logger,
              torch.nn.Module,
              object,
              metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        # for some obscure reason, calling the super init does not lead to the call of both classes.
        # by consequence, we make the calls manually to each parents
        torch.nn.Module.__init__(self)
        _Logger.__init__(self, *args, **kwargs)

    def __repr__(self):
        return capitalize_only_first(self.__str__())

    def set_log_level(self, level: int = None) -> int:
        level = super(_Module, self).set_log_level(level)
        for child in self.children():
            if isinstance(child, _Logger):
                child.set_log_level(level)
        return level

    def _euclidean_parameters(self, recurse=True):
        if recurse:
            for module in self.children():
                if isinstance(module, _Module):
                    yield from module._euclidean_parameters(recurse)

    def _stiefel_parameters(self, recurse=True):
        if recurse:
            for module in self.children():
                if isinstance(module, _Module):
                    yield from module._stiefel_parameters(recurse)

    def _slow_parameters(self, recurse=True):
        if recurse:
            for module in self.children():
                if isinstance(module, _Module):
                    yield from module._slow_parameters(recurse)

    def manifold_parameters(self, recurse=True, type='euclidean') -> Iterator[torch.nn.Parameter]:
        switcher = {'euclidean': self._euclidean_parameters,
                    'stiefel': self._stiefel_parameters,
                    'slow': self._slow_parameters}
        gen = switcher.get(type, 'Invalid manifold name.')

        memo = set()
        for p in gen(recurse=recurse):
            if p not in memo:
                memo.add(p)
                yield p



# coding=utf-8
from functools import wraps


def kwargs_decorator(dict_kwargs):
    r"""
    Decorator that appends the dictionnary of method keyword arguments (kwargs) with additional values provided in
    `dict_kwargs`. If a keyword already exists in the original dictionnary, its value is neglected. This is useful for
    adding default values to the keyword arguments of methods, e.g., constructors. An alternative is to use the
    `kwargs.pop(keyword, default)` instead.
    
    :param dict_kwargs: Dictionnary to be appended to the original dictionary. Redundant keywords are neglected.
    :return: Appended dictionnary.

    :type dict_kwargs: dict
    :rtype: dict
    """
    def wrapper(f):
        @wraps(f)
        def inner_wrapper(*args, **kwargs):
            new_kwargs = {**dict_kwargs, **kwargs}
            return f(*args, **new_kwargs)

        return inner_wrapper

    return wrapper


class extend_docstring:
    r"""
    Decorator adding the documentation of the provided definition to the decorated definition. This useful for
    inheriting the documentation.
    """
    def __init__(self, method):
        self.doc = method.__doc__

    def __call__(self, function):
        if self.doc is not None:
            doc = function.__doc__
            function.__doc__ = self.doc
            if doc is not None:
                function.__doc__ = doc + function.__doc__
        return function



# coding=utf-8
from __future__ import annotations
import nested_dict

from .Model import Model
from ..utils import capitalize_only_first, KerchError


class RKM(Model):
    def __init__(self, *args, **kwargs):
        super(RKM, self).__init__(*args, **kwargs)
        level = 0
        while True:
            try:
                level_kwargs = kwargs[f"level{level}"]
                self.append_level(**level_kwargs)
                level += 1
            except KeyError:
                break
        if level == 0:
           raise KerchError(cls=self, message='The RKM model does not contain any level. '
                                              'Please add one or check if the arguments are correct.')

    def __str__(self):
        description = f"[Model] RKM with the following levels:"
        nl, tab = '\n', '\t'
        for num, level in enumerate(self.levels):
            dim_input = level.dim_input
            dim_output = level.dim_output
            description += f"{nl}{tab}* L{num}[{dim_input},{dim_output}]: " \
                           f"{capitalize_only_first(level.representation)} {level}"
        return description

    def append_level(self, *args, **kwargs) -> None:
        self._append_level(*args, **kwargs)







# coding=utf-8
from __future__ import annotations

from .Watcher import Watcher


def factory(watcher_type:str='saver', **kwargs) -> Watcher | None:
    r"""
    # TODO
    """

    watcher = class_factory(watcher_type)
    return watcher(**kwargs)


def class_factory(watcher_type:str='saver'):
    if watcher_type == 'none':
        return None

    def case_insensitive_getattr(obj, attr):
        for a in dir(obj):
            if a.lower() == attr.lower():
                return getattr(obj, a)
        return None

    import kerch.monitor
    watcher_class = case_insensitive_getattr(kerch.watch, watcher_type)
    if watcher_class is None:
        raise NameError("Invalid watcher type.")
    return watcher_class




# coding=utf-8
from .Model import Model

def factory(model_type:str = 'rkm', **kwargs) -> Model:
    r"""
    Creates a kernel based on the specified name with the specified arguments. This is the same as
    calling `kerch.kernel.name(*args, **kwargs)` (if `name` is not a string here). This allows for the creation of kernel where
    the name of kernel is passed as a string.

    :param kernel_type: Type of kernel chosen. For the possible choices, please refer to the (non-abstract) classes
        herebelow. Defaults to kerch.DEFAULT_KERNEL_TYPE.
    :param \**kwargs: Arguments to be passed to the kernel constructor, such as `sample` or `sigma`. If an argument is
        passed that does not exist (e.g. `sigma` to a `linear` kernel), it will just be neglected. For the default
        values, please refer to the default values of the requested kernel.
    :type kernel_type: str, optional
    :type \**kwargs: dict, optional
    :return: An instantiation of the specified kernel.
    """

    model = class_factory(model_type)
    return model(**kwargs)

def class_factory(model_type: str = 'rkm'):
    def case_insensitive_getattr(obj, attr):
        for a in dir(obj):
            if a.lower() == attr.lower():
                return getattr(obj, a)
        return None

    import kerch.model
    model = case_insensitive_getattr(kerch.model, model_type)
    if model is None:
        raise NameError("Invalid model type.")
    return model




# coding=utf-8
from .logger import Logger as Logger
from .module import Module as Module
from .cache import Cache as Cache
from .stochastic import Stochastic as Stochastic
from .sample import Sample as Sample
